/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

void alloc_loglike(void)
{
  loglike = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);
  lambda = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);

  if (mode_smooth)
    log_prior = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);

  loglike_best =
    (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);
  lambda_best = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);

  if (mode_smooth)
    log_prior_best = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);
}

void alloc_log_itemp(void)
{
  log_itemp = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);
  log_itemp_best = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*ll_size);
}

void expand_loglike_all(void)
{
  int old_ll_size;

  old_ll_size = ll_size;
  ll_size *= 2;

  loglike =
    (FLOAT_T *)xrealloc(loglike,sizeof(FLOAT_T)*ll_size);
  lambda = (FLOAT_T *)xrealloc(lambda,sizeof(FLOAT_T)*ll_size);

  if (mode_smooth)
    log_prior =
      (FLOAT_T *)xrealloc(log_prior,sizeof(FLOAT_T)*ll_size);

  loglike_best =
    (FLOAT_T *)xrealloc(loglike_best,sizeof(FLOAT_T)*ll_size);
  lambda_best = (FLOAT_T *)xrealloc(lambda_best,sizeof(FLOAT_T)*ll_size);

  if (mode_smooth)
    log_prior_best =
      (FLOAT_T *)xrealloc(log_prior_best,sizeof(FLOAT_T)*ll_size);

  if (mode_daem) {
    log_itemp = (FLOAT_T *)xrealloc(log_itemp,sizeof(FLOAT_T)*ll_size);
    log_itemp_best =
      (FLOAT_T *)xrealloc(log_itemp_best,sizeof(FLOAT_T)*ll_size);
  }
}

void alloc_prob(void)
{
  int k;

  prob_comp = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);
  prob_cond = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++) {
    prob_comp[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
    prob_cond[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
  }

  prob_incomp = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
}

void alloc_ic(int num_search)
{
  ic = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_search);
  bic = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_search);
  autoclass = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_search);
}

void alloc_seed(void)
{
  int u,num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  log_seed = (unsigned int **)xmalloc(sizeof(unsigned int *)*num_search);
  log_num_iterate = (int **)xmalloc(sizeof(int *)*num_search);
  log_final_lambda = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_search);
  log_r_best = (int *)xmalloc(sizeof(int)*num_search);

  for (u = 0; u < num_search; u++) {
    log_seed[u] = (unsigned int *)xmalloc(sizeof(unsigned int)*num_init);
    log_num_iterate[u] = (int *)xmalloc(sizeof(int)*num_init);
    log_final_lambda[u] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_init);
  }
}

void release_prob(void)
{
  int k;
  for (k=0; k<num_class; k++) {
    xfree(prob_comp[k]);
    xfree(prob_cond[k]);
  }
  xfree(prob_comp);
  xfree(prob_incomp);
  xfree(prob_cond);
}

void release_loglike(void)
{
  xfree(loglike);
  xfree(lambda);
  if (mode_smooth) xfree(log_prior);

  xfree(loglike_best);
  xfree(lambda_best);
  if (mode_smooth) xfree(log_prior_best);
}

void release_log_itemp(void)
{
  xfree(log_itemp);
  xfree(log_itemp_best);
}

void release_ic(void)
{
  xfree(ic);
  xfree(bic);
  xfree(autoclass);
}

void release_seed(void)
{
  int u, num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  for (u=0; u < num_search; u++) {
    xfree(log_seed[u]);
    xfree(log_num_iterate[u]);
    xfree(log_final_lambda[u]);
  }
  xfree(log_seed);
  xfree(log_num_iterate);
  xfree(log_final_lambda);
}

void alloc_param(void)
{
  int k,v,u,n;

  param_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);
    for (k=0; k<num_class; k++) {
      param_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        param_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    param_u = (VECTOR **)xmalloc(sizeof(MATRIX *)*num_class);
    param_s = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);
    param_p = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);
    param_sdet = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u[k] = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      param_s[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      param_p[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      param_sdet[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        param_u[k][u] = alloc_vector(n);
        param_s[k][u] = alloc_matrix(n,n);
        param_p[k][u] = alloc_matrix(n,n);
      }
    }
  }

}

void alloc_param_best(void)
{
  int k,v,u,n;

  param_c_best = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w_best = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      param_w_best[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        param_w_best[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    param_u_best = (VECTOR **)xmalloc(sizeof(VECTOR *)*num_class);
    param_s_best = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_best[k] = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      param_s_best[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        param_u_best[k][u] = alloc_vector(n);
        param_s_best[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void save_param_best(void)
{
  int k,v,j,u;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c_best[k] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (j=0; j<num_word[v]; j++)
          param_w_best[k][v][j] = param_w[k][v][j];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(param_u_best[k][u],param_u[k][u]);
        copy_matrix(param_s_best[k][u],param_s[k][u]);
      }
    }
  }
}

void restore_param_best(void)
{
  int k,v,j,u;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c[k] = param_c_best[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (j=0; j<num_word[v]; j++)
          param_w[k][v][j] = param_w_best[k][v][j];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(param_u[k][u],param_u_best[k][u]);
        copy_matrix(param_s[k][u],param_s_best[k][u]);
        param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
        if (err_inv) DIE_NUMERIC("Failed to invert covariances\n");
      }
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void save_param_opt(void)
{
  int k,v,i,u,n;

  param_c_opt = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class_opt);

  if (mode_nominal) {
    param_w_opt = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class_opt);

    for (k=0; k<num_class_opt; k++) {
      param_w_opt[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        param_w_opt[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class_opt; k++)
    param_c_opt[k] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class_opt; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          param_w_opt[k][v][i] = param_w[k][v][i];
  }

  if (mode_gauss) {
    param_u_opt = (VECTOR **)xmalloc(sizeof(VECTOR *)*num_class);
    param_s_opt = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_opt[k] = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      param_s_opt[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        param_u_opt[k][u] = alloc_vector(n);
        param_s_opt[k][u] = alloc_matrix(n,n);
      }
    }

#pragma omp parallel for private(u)
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(param_u_opt[k][u],param_u[k][u]);
        copy_matrix(param_s_opt[k][u],param_s[k][u]);
      }
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void restore_param_opt(void)
{
  int k,v,i,u,n;

  param_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class_opt);

  if (mode_nominal) {
    param_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class_opt);

    for (k=0; k<num_class_opt; k++) {
      param_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        param_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class_opt; k++)
    param_c[k] = param_c_opt[k];

  if (mode_nominal) {
#pragma omp parallel for private(i,v)
    for (k=0; k<num_class_opt; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          param_w[k][v][i] = param_w_opt[k][v][i];
  }

  if (mode_gauss) {
    param_u = (VECTOR **)xmalloc(sizeof(VECTOR *)*num_class_opt);
    param_s = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class_opt);
    param_p = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class_opt);
    param_sdet = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class_opt);

    for (k=0; k<num_class_opt; k++) {
      param_u[k] = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      param_s[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      param_p[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      param_sdet[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        param_u[k][u] = alloc_vector(n);
        param_s[k][u] = alloc_matrix(n,n);
        param_p[k][u] = alloc_matrix(n,n);
      }
    }

#pragma omp parallel for private(u)
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(param_u[k][u],param_u_opt[k][u]);
        copy_matrix(param_s[k][u],param_s_opt[k][u]);
        param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
        if (err_inv) DIE_NUMERIC("Failed to invert covariances\n");
      }
    }
  }

  if (mode_gauss) alloc_hparam_gauss();
}

void alloc_hparam(void)
{
  int k,v;

  hparam_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);
    for (k=0; k<num_class; k++) {
      hparam_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        hparam_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }
}

void release_hparam(void)
{
  int k,v;

  xfree(hparam_c);

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(hparam_w[k][v]);
      xfree(hparam_w[k]);
    }
    xfree(hparam_w);
  }
}

void alloc_hparam_gauss(void)
{
  int k,u,n;

  /* A bit dirty.. */
#ifdef VBNC
  if (mode_gauss && mode_smooth_gauss) {
#else
  if (mode_gauss) {
#endif

    hparam_u = (VECTOR **)xmalloc(sizeof(MATRIX *)*num_class);
    hparam_s = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);
    hparam_p = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);
    hparam_sdet = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_u[k] = (VECTOR *)xmalloc(sizeof(MATRIX)*gauss_size);
      hparam_s[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      hparam_p[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        hparam_u[k][u] = alloc_vector(n);
        hparam_s[k][u] = alloc_matrix(n,n);
        hparam_p[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void release_hparam_gauss(void)
{
  int k,u;

  if (mode_gauss && mode_smooth_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(hparam_u[k][u]);
        release_matrix(hparam_s[k][u]);
        release_matrix(hparam_p[k][u]);
      }
      xfree(hparam_u[k]);
      xfree(hparam_s[k]);
      xfree(hparam_p[k]);
      xfree(hparam_sdet[k]);
    }
    xfree(hparam_u);
    xfree(hparam_s);
    xfree(hparam_p);
    xfree(hparam_sdet);
  }
}

void alloc_hparam_gauss_aux(void)
{
  int k;

  hparam_tau = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);
  hparam_alpha_w = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);

  for (k=0; k<num_class; k++)
    hparam_alpha_w[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_size);
}

void release_hparam_gauss_aux(void)
{
  int k;

  for (k=0; k<num_class; k++)
    xfree(hparam_alpha_w[k]);

  xfree(hparam_tau);
  xfree(hparam_alpha_w);
}

void alloc_expect(void)
{
  int k,v,u,n;

  expect_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    expect_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      expect_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        expect_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    expect_u = (VECTOR **)xmalloc(sizeof(VECTOR *)*num_class);
    expect_s = (MATRIX **)xmalloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      expect_u[k] = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      expect_s[k] = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        expect_u[k][u] = alloc_vector(n);
        expect_s[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void alloc_prob_mem(void)
{
  int k,v,u;

  prob_mem_case = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++)
    prob_mem_case[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);

  prob_compo_case = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++)
    prob_compo_case[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);

  prob_case = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);

  if (mode_nominal) {
    prob_mem_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);
    for (k=0; k<num_class; k++) {
      prob_mem_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        prob_mem_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }

    prob_compo_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);
    for (k=0; k<num_class; k++) {
      prob_compo_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        prob_compo_w[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }

    prob_w = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      prob_w[v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
  }

  if (mode_gauss) {
    prob_mem_g = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      prob_mem_g[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        prob_mem_g[k][u] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
    }

    prob_compo_g = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      prob_compo_g[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        prob_compo_g[k][u] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
    }

    prob_g = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      prob_g[u] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
  }
}

void release_param(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(param_w[k][v]);
      xfree(param_w[k]);
    }
    xfree(param_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(param_u[k][u]);
        release_matrix(param_s[k][u]);
        release_matrix(param_p[k][u]);
      }
      xfree(param_u[k]);
      xfree(param_s[k]);
      xfree(param_p[k]);
      xfree(param_sdet[k]);
    }
    xfree(param_u);
    xfree(param_s);
    xfree(param_p);
    xfree(param_sdet);
  }

  xfree(param_c);
}

void release_param_best(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(param_w_best[k][v]);
      xfree(param_w_best[k]);
    }
    xfree(param_w_best);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(param_u_best[k][u]);
        release_matrix(param_s_best[k][u]);
      }
      xfree(param_u_best[k]);
      xfree(param_s_best[k]);
    }
    xfree(param_u_best);
    xfree(param_s_best);
  }

  xfree(param_c_best);
}

void release_param_opt(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class_opt; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(param_w_opt[k][v]);
      xfree(param_w_opt[k]);
    }
    xfree(param_w_opt);
  }

  if (mode_gauss) {
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(param_u_opt[k][u]);
        release_matrix(param_s_opt[k][u]);
      }
      xfree(param_u_opt[k]);
      xfree(param_s_opt[k]);
    }
    xfree(param_u_opt);
    xfree(param_s_opt);
  }

  xfree(param_c_opt);
}

void release_expect(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(expect_w[k][v]);
      xfree(expect_w[k]);
    }
    xfree(expect_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(expect_u[k][u]);
        release_matrix(expect_s[k][u]);
      }
      xfree(expect_u[k]);
      xfree(expect_s[k]);
    }
    xfree(expect_u);
    xfree(expect_s);
  }

  xfree(expect_c);
}

void release_prob_mem(void)
{
  int k,v,u;

  for (k=0; k<num_class; k++) {
    xfree(prob_compo_case[k]);
    xfree(prob_mem_case[k]);
  }
  xfree(prob_mem_case);
  xfree(prob_compo_case);
  xfree(prob_case);

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        xfree(prob_mem_w[k][v]);
        xfree(prob_compo_w[k][v]);
      }
      xfree(prob_mem_w[k]);
      xfree(prob_compo_w[k]);
    }
    xfree(prob_mem_w);
    xfree(prob_compo_w);

    for (v=0; v<tuple_size_d; v++)
      xfree(prob_w[v]);
    xfree(prob_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        xfree(prob_mem_g[k][u]);
        xfree(prob_compo_g[k][u]);
      }
      xfree(prob_mem_g[k]);
      xfree(prob_compo_g[k]);
    }
    xfree(prob_mem_g);
    xfree(prob_compo_g);

    for (u=0; u<gauss_size; u++)
      xfree(prob_g[u]);
    xfree(prob_g);
  }
}

void alloc_read_param(int max_num_class, int *max_num_word)
{
  int k,v,u,a;

  read_param_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*max_num_class);

  if (mode_nominal) {
    read_param_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_param_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        read_param_w[k][v] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*max_num_word[v]);
    }
  }

  if (mode_gauss) {
    read_param_u = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*max_num_class);
    read_param_s = (FLOAT_T ****)xmalloc(sizeof(FLOAT_T ***)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_param_u[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      read_param_s[k] = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        read_param_u[k][u] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
        read_param_s[k][u] =
          (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);

        for (a=0; a<gauss_attr_size[u]; a++)
          read_param_s[k][u][a] =
            (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
  }
}

void release_read_param(int max_num_class)
{
  int k,v,u,a;

  xfree(read_param_c);

  if (mode_nominal) {
    for (k=0; k<max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(read_param_w[k][v]);
      xfree(read_param_w[k]);
    }
    xfree(read_param_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        xfree(read_param_u[k][u]);
        for (a=0; a<gauss_attr_size[u]; a++)
          xfree(read_param_s[k][u][a]);
      }
      xfree(read_param_u[k]);
      xfree(read_param_s[k]);
    }
    xfree(read_param_u);
    xfree(read_param_s);
  }
}

void expand_read_param_c(int old_max_num_class, int new_max_num_class)
{
  int i;
  FLOAT_T *new_read_param_c;

  new_read_param_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*new_max_num_class);

  for (i=0; i<old_max_num_class; i++)
    new_read_param_c[i] = read_param_c[i];

  xfree(read_param_c);
  read_param_c = new_read_param_c;
}

void expand_read_param_w(int old_max_num_class, int new_max_num_class,
                         int *old_max_num_word, int *new_max_num_word)
{
  int k,i,v;
  FLOAT_T ***new_read_param_w;

  if (mode_nominal) {
    new_read_param_w =
      (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*new_max_num_class);

    for (k=0; k<new_max_num_class; k++) {
      new_read_param_w[k] =
        (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++)
        new_read_param_w[k][v] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*new_max_num_word[v]);
    }

    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        for (i=0; i<old_max_num_word[v]; i++) {
          new_read_param_w[k][v][i] = read_param_w[k][v][i];
        }
      }
    }

    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(read_param_w[k][v]);
      xfree(read_param_w[k]);
    }
    xfree(read_param_w);

    read_param_w = new_read_param_w;
  }
}

void expand_read_param_u(int old_max_num_class, int new_max_num_class)
{
  int k,u,a;
  FLOAT_T ***new_read_param_u;

  if (mode_gauss) {
    new_read_param_u =
      (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*new_max_num_class);

    for (k=0; k<new_max_num_class; k++) {
      new_read_param_u[k] =
        (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        new_read_param_u[k][u] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
    }

    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (a=0; a<gauss_attr_size[u]; a++)
          new_read_param_u[k][u][a] = read_param_u[k][u][a];

    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++)
        xfree(read_param_u[k][u]);
      xfree(read_param_u[k]);
    }
    xfree(read_param_u);

    read_param_u = new_read_param_u;
  }
}

void expand_read_param_s(int old_max_num_class, int new_max_num_class)
{
  int k,u,a,a0,a1;
  FLOAT_T ****new_read_param_s;

  if (mode_gauss) {
    new_read_param_s =
      (FLOAT_T ****)xmalloc(sizeof(FLOAT_T ***)*new_max_num_class);

    for (k=0; k<new_max_num_class; k++) {
      new_read_param_s[k] =
        (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        new_read_param_s[k][u] =
          (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
        for (a=0; a<gauss_attr_size[u]; a++)
          new_read_param_s[k][u][a] =
            (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }

    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (a0=0; a0<gauss_attr_size[u]; a0++)
          for (a1=0; a1<gauss_attr_size[u]; a1++)
            new_read_param_s[k][u][a0][a1] = read_param_s[k][u][a0][a1];

    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        for (a=0; a<gauss_attr_size[u]; a++)
          xfree(read_param_s[k][u][a]);
        xfree(read_param_s[k][u]);
      }
      xfree(read_param_s[k]);
    }
    xfree(read_param_s);

    read_param_s = new_read_param_s;
  }
}

void alloc_read_hparam(int max_num_class, int *max_num_word)
{
  int k,v,u,a;

  read_hparam_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*max_num_class);

  if (mode_nominal) {
    read_hparam_w = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_hparam_w[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        read_hparam_w[k][v] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*max_num_word[v]);
    }
  }

  if (mode_gauss) {
    read_hparam_u = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*max_num_class);
    read_hparam_s = (FLOAT_T ****)xmalloc(sizeof(FLOAT_T ***)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_hparam_u[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      read_hparam_s[k] = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        read_hparam_u[k][u] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
        read_hparam_s[k][u] =
          (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);

        for (a=0; a<gauss_attr_size[u]; a++)
          read_hparam_s[k][u][a] =
            (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }

    read_tau = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*max_num_class);
    read_alpha_w = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*max_num_class);
    for (k=0; k<max_num_class; k++)
      read_alpha_w[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_size);
  }
}

void release_read_hparam(int max_num_class)
{
  int k,v,u,a;

  xfree(read_hparam_c);

  if (mode_nominal) {
    for (k=0; k<max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(read_hparam_w[k][v]);
      xfree(read_hparam_w[k]);
    }
    xfree(read_hparam_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        xfree(read_hparam_u[k][u]);
        for (a=0; a<gauss_attr_size[u]; a++)
          xfree(read_hparam_s[k][u][a]);
      }
      xfree(read_hparam_u[k]);
      xfree(read_hparam_s[k]);
    }
    xfree(read_hparam_u);
    xfree(read_hparam_s);

    for (k=0; k<num_class; k++)
      xfree(read_alpha_w[k]);

    xfree(read_tau);
    xfree(read_alpha_w);
  }
}

void expand_read_hparam_c(int old_max_num_class, int new_max_num_class)
{
  int i;
  FLOAT_T *new_read_hparam_c;

  new_read_hparam_c = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*new_max_num_class);

  for (i=0; i<old_max_num_class; i++)
    new_read_hparam_c[i] = read_hparam_c[i];

  xfree(read_hparam_c);
  read_hparam_c = new_read_hparam_c;
}

void expand_read_hparam_w(int old_max_num_class, int new_max_num_class,
                          int *old_max_num_word, int *new_max_num_word)
{
  int k,i,v;
  FLOAT_T ***new_read_hparam_w;

  if (mode_nominal) {
    new_read_hparam_w =
      (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*new_max_num_class);
    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_w[k] =
        (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        new_read_hparam_w[k][v] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*new_max_num_word[v]);
    }

    for (k=0; k<old_max_num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<old_max_num_word[v]; i++)
          new_read_hparam_w[k][v][i] = read_hparam_w[k][v][i];

    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(read_hparam_w[k][v]);
      xfree(read_hparam_w[k]);
    }
    xfree(read_hparam_w);

    read_hparam_w = new_read_hparam_w;
  }
}

void expand_read_hparam_u(int old_max_num_class, int new_max_num_class)
{
  int k,u,a;
  FLOAT_T ***new_read_hparam_u;

  if (mode_gauss) {
    new_read_hparam_u =
      (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*new_max_num_class);

    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_u[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        new_read_hparam_u[k][u] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
    }

    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (a=0; a<gauss_attr_size[u]; a++)
          new_read_hparam_u[k][u][a] = read_hparam_u[k][u][a];

    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++)
        xfree(read_hparam_u[k][u]);
      xfree(read_hparam_u[k]);
    }
    xfree(read_hparam_u);

    read_hparam_u = new_read_hparam_u;
  }
}

void expand_read_hparam_s(int old_max_num_class, int new_max_num_class)
{
  int k,u,a,a0,a1;
  FLOAT_T ****new_read_hparam_s;

  if (mode_gauss) {
    new_read_hparam_s =
      (FLOAT_T ****)xmalloc(sizeof(FLOAT_T ***)*new_max_num_class);

    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_s[k] =
        (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        new_read_hparam_s[k][u] =
          (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
        for (a=0; a<gauss_attr_size[u]; a++)
          new_read_hparam_s[k][u][a] =
            (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }

    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (a0=0; a0<gauss_attr_size[u]; a0++)
          for (a1=0; a1<gauss_attr_size[u]; a1++)
            new_read_hparam_s[k][u][a0][a1] = read_hparam_s[k][u][a0][a1];

    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        for (a=0; a<gauss_attr_size[u]; a++)
          xfree(read_hparam_s[k][u][a]);
        xfree(read_hparam_s[k][u]);
      }
      xfree(read_hparam_s[k]);
    }
    xfree(read_hparam_s);

    read_hparam_s = new_read_hparam_s;
  }
}

void alloc_cluster(void)
{
  int k,v,u;

  if (mode_nominal) {
    cluster_w_pred = (int **)xmalloc(sizeof(int *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      cluster_w_pred[v] = (int *)xmalloc(sizeof(int)*num_word[v]);

    cluster_w_freq = (int **)xmalloc(sizeof(int *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      cluster_w_freq[v] = (int *)xmalloc(sizeof(int)*num_class);

    cluster_w = (int ***)xmalloc(sizeof(int **)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++) {
      cluster_w[v] = (int **)xmalloc(sizeof(int *)*num_class);
      for (k=0; k<num_class; k++)
        cluster_w[v][k] = (int *)xmalloc(sizeof(int)*num_word[v]);
    }
  }

  if (mode_gauss) {
    cluster_g_pred = (int **)xmalloc(sizeof(int *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      cluster_g_pred[u] = (int *)xmalloc(sizeof(int)*num_entry);

    cluster_g_freq = (int **)xmalloc(sizeof(int *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      cluster_g_freq[u] = (int *)xmalloc(sizeof(int)*num_class);

    cluster_g = (int ***)xmalloc(sizeof(int **)*gauss_size);
    for (u=0; u<gauss_size; u++) {
      cluster_g[u] = (int **)xmalloc(sizeof(int *)*num_class);
      for (k=0; k<num_class; k++)
        cluster_g[u][k] = (int *)xmalloc(sizeof(int)*num_entry);
    }
  }

  /* redundant, but safe */
  cluster_case_pred = (int *)xmalloc(sizeof(int)*num_entry);
  cluster_case_freq = (int *)xmalloc(sizeof(int)*num_class);

  cluster_case = (int **)xmalloc(sizeof(int *)*num_class);
  for (k=0; k<num_class; k++)
    cluster_case[k] =
      (int *)xmalloc(sizeof(int)*num_entry);  /* redundant, but safe */
}

void release_cluster(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++) {
      for (k=0; k<num_class; k++)
        xfree(cluster_w[v][k]);
      xfree(cluster_w[v]);
    }
    xfree(cluster_w);

    for (v=0; v<tuple_size_d; v++) {
      xfree(cluster_w_pred[v]);
      xfree(cluster_w_freq[v]);
    }
    xfree(cluster_w_pred);
    xfree(cluster_w_freq);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      for (k=0; k<num_class; k++)
        xfree(cluster_g[u][k]);
      xfree(cluster_g[u]);
    }
    xfree(cluster_g);

    for (u=0; u<gauss_size; u++) {
      xfree(cluster_g_pred[u]);
      xfree(cluster_g_freq[u]);
    }
    xfree(cluster_g_pred);
    xfree(cluster_g_freq);
  }

  for (k=0; k<num_class; k++)
    xfree(cluster_case[k]);
  xfree(cluster_case);

  xfree(cluster_case_pred);
  xfree(cluster_case_freq);
}

void alloc_rank(void)
{
  int k,v,u;

  rank_class = (int *)xmalloc(sizeof(int)*num_class);
  rank_mem_case = (int **)xmalloc(sizeof(int **)*num_class);

  for (k=0; k<num_class; k++)
    rank_mem_case[k] = (int *)xmalloc(sizeof(int)*num_entry);

  if (mode_nominal) {
    rank_mem_w = (int ***)xmalloc(sizeof(int **)*num_class);

    for (k=0; k<num_class; k++) {
      rank_mem_w[k] = (int **)xmalloc(sizeof(int *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        rank_mem_w[k][v] = (int *)xmalloc(sizeof(int)*num_word[v]);
    }
  }

  if (mode_gauss) {
    rank_mem_g = (int ***)xmalloc(sizeof(int **)*num_class);

    for (k=0; k<num_class; k++) {
      rank_mem_g[k] = (int **)xmalloc(sizeof(int *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        rank_mem_g[k][u] = (int *)xmalloc(sizeof(int)*num_entry);
    }
  }
}

void release_rank(void)
{
  int k,v,u;

  xfree(rank_class);

  for (k=0; k<num_class; k++)
    xfree(rank_mem_case[k]);
  xfree(rank_mem_case);

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(rank_mem_w[k][v]);
      xfree(rank_mem_w[k]);
    }
    xfree(rank_mem_w);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++)
        xfree(rank_mem_g[k][u]);
      xfree(rank_mem_g[k]);
    }
    xfree(rank_mem_g);
  }
}

void alloc_ll_sm_size(void)
{
  int t;

  ll_sm_size = (int *)xmalloc(sizeof(int)*(num_sm + 1));

  for (t = 0; t < (num_sm + 1); t++) {
    ll_sm_size[t] = MAX_MAX_ITERATE;
  }
}

void alloc_loglike_sm(void)
{
  int n,t;

  /* We need one extra entry to record the values for the first full EM */
  loglike_sm = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*(num_sm+1));
  log_prior_sm = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*(num_sm+1));
  lambda_sm = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*(num_sm+1));

  for (t=0; t<(num_sm+1); t++) {  /* Be careful for the terminal condition */
    n = ll_sm_size[t];
    /* At maximum, we have n partial EM iterations and n full EM iterations: */
    loglike_sm[t] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*(2*n));
    log_prior_sm[t] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*(2*n));
    lambda_sm[t] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*(2*n));
  }
}

void alloc_log_itemp_sm(void)
{
  int n,t;

  log_itemp_sm = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*(num_sm+1));

  for (t=0; t<(num_sm+1); t++) {
    n = ll_sm_size[t];
    log_itemp_sm[t] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*(2*n));
  }
}

void expand_loglike_sm_all(int t)
{
  int old_ll_size;

  old_ll_size = ll_sm_size[t];
  ll_sm_size[t] *= 2;

  loglike_sm[t] = (FLOAT_T *)xrealloc(loglike_sm[t],
                                      sizeof(FLOAT_T)*ll_sm_size[t]);
  lambda_sm[t] = (FLOAT_T *)xrealloc(lambda_sm[t],
                                     sizeof(FLOAT_T)*ll_sm_size[t]);

  if (mode_smooth)
    log_prior_sm[t] = (FLOAT_T *)xrealloc(log_prior_sm[t],
                                          sizeof(FLOAT_T)*ll_sm_size[t]);

  if (mode_daem)
    log_itemp_sm[t] = (FLOAT_T *)xrealloc(log_itemp_sm[t],
                                          sizeof(FLOAT_T)*ll_sm_size[t]);

  loglike_best = (FLOAT_T *)xrealloc(loglike_best,sizeof(FLOAT_T)*ll_size);
  lambda_best = (FLOAT_T *)xrealloc(lambda_best,sizeof(FLOAT_T)*ll_size);

  if (mode_smooth)
    log_prior_best =
      (FLOAT_T *)xrealloc(log_prior_best,sizeof(FLOAT_T)*ll_size);

  if (mode_daem) {
    log_itemp = (FLOAT_T *)xrealloc(log_itemp,sizeof(FLOAT_T)*ll_size);
    log_itemp_best =
      (FLOAT_T *)xrealloc(log_itemp_best,sizeof(FLOAT_T)*ll_size);
  }
}

void alloc_jmerge()
{
  int k;

  jmerge = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++)
    jmerge[k] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);
}

void alloc_jsplit()
{
  jsplit = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);
}

void alloc_candidate()
{
  int h;

  sm_candidate = (int **)xmalloc(sizeof(int *)*num_candidate);
  for (h=0; h<num_candidate; h++)
    sm_candidate[h] = (int *)xmalloc(sizeof(int)*3);
}

void alloc_prob_part(void)
{
  int l;

  prob_incomp_part = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);

  prob_cond_part = (FLOAT_T **)xmalloc(sizeof(FLOAT_T*)*3);
  for (l=0; l<3; l++)
    prob_cond_part[l] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_entry);
}

void alloc_num_iterate_sm()
{
  num_iterate_sm_full = (int *)xmalloc(sizeof(int)*(num_sm+1));
  num_iterate_sm_part = (int *)xmalloc(sizeof(int)*(num_sm+1));
}

void alloc_num_result_sm()
{
  num_result_rejected = (int *)xmalloc(sizeof(int)*num_sm);
  num_result_skipped = (int *)xmalloc(sizeof(int)*num_sm);
}

void release_jmerge()
{
  int k;
  for (k=0; k<num_class; k++)
    xfree(jmerge[k]);
  xfree(jmerge);
}

void release_jsplit()
{
  xfree(jsplit);
}

void release_candidate()
{
  int h;

  for (h=0; h<num_candidate; h++)
    xfree(sm_candidate[h]);
  xfree(sm_candidate);
}

void release_prob_part(void)
{
  int l;

  for (l=0; l<3; l++)
    xfree(prob_cond_part[l]);
  xfree(prob_cond_part);
  xfree(prob_incomp_part);
}

void release_ll_sm_size(void)
{
  xfree(ll_sm_size);
}

void release_loglike_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    xfree(loglike_sm[t]);
    xfree(log_prior_sm[t]);
    xfree(lambda_sm[t]);
  }
  xfree(loglike_sm);
  xfree(log_prior_sm);
  xfree(lambda_sm);
}

void release_log_itemp_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++)
    xfree(log_itemp_sm[t]);
  xfree(log_itemp_sm);
}

void release_num_iterate_sm()
{
  xfree(num_iterate_sm_full);
  xfree(num_iterate_sm_part);
}

void release_num_result_sm()
{
  xfree(num_result_rejected);
  xfree(num_result_skipped);
}

void alloc_param_part(void)
{
  param_c_part = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*3);
}

void  release_param_part(void)
{
  xfree(param_c_part);
}

void alloc_param_sm(void)
{
  int k,v,u,a;

  param_c_sm = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w_sm = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      param_w_sm[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        param_w_sm[k][v] = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    param_u_sm = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*num_class);
    param_s_sm = (FLOAT_T ****)xmalloc(sizeof(FLOAT_T ***)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_sm[k] = (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_size);
      param_s_sm[k] = (FLOAT_T ***)xmalloc(sizeof(FLOAT_T **)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        param_u_sm[k][u] =
          (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
        param_s_sm[k][u] =
          (FLOAT_T **)xmalloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);

        for (a=0; a<gauss_attr_size[u]; a++)
          param_s_sm[k][u][a] =
            (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
  }
}

void release_param_sm(void)
{
  int k,v,u,a;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        xfree(param_w_sm[k][v]);
      xfree(param_w_sm[k]);
    }
    xfree(param_w_sm);
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        xfree(param_u_sm[k][u]);
        for (a=0; a<gauss_attr_size[u]; a++)
          xfree(param_s_sm[k][u][a]);
      }
      xfree(param_u_sm[k]);
      xfree(param_s_sm[k]);
    }
    xfree(param_u_sm);
    xfree(param_s_sm);
  }

  xfree(param_c_sm);
}

void alloc_eval(void)
{
  int k;

  cluster_case_freq_joint = (int **)xmalloc(sizeof(int *)*num_class);
  for (k=0; k<num_class; k++)
    cluster_case_freq_joint[k] = (int *)xmalloc(sizeof(int)*num_ans_class);
}

void release_eval(void)
{
  int k;

  for (k=0; k<num_class; k++)
    xfree(cluster_case_freq_joint[k]);
  xfree(cluster_case_freq_joint);
}
