/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void compute_cluster_w_ml(void)
{
  int *tmp_count;
  int k,i,v;
  int k_ml;
  FLOAT_T p_max;

  /* Allocate the temporary areas */
  tmp_count = (int *)xmalloc(sizeof(int)*num_class);

  for (v=0; v<tuple_size_d; v++)
    for (k=0; k<num_class; k++)
      cluster_w_freq[v][k] = 0;

  for (v=0; v<tuple_size_d; v++) {
    for (i=0; i<num_word[v]; i++) {
      k_ml = 0;
      p_max = 0.0;
      for (k=0; k<num_class; k++) {
        if (prob_mem_w[k][v][i] > p_max) {
          k_ml = k;
          p_max = prob_mem_w[k][v][i];
        }
      }
      cluster_w_pred[v][i] = k_ml;
      cluster_w_freq[v][k_ml]++; /* Count up att-vals in class k_ml */
    }
  }

  for (v=0; v<tuple_size_d; v++) {
    for (k=0; k<num_class; k++)
      tmp_count[k] = 0;

    for (i=0; i<num_word[v]; i++) {
      k_ml = cluster_w_pred[v][i];
      cluster_w[v][k_ml][tmp_count[k_ml]] = i;
      tmp_count[k_ml]++;
    }

    if (mode_lex_order) {
      tmp_word_list = word_list[v];
      for (k=0; k<num_class; k++) {
        qsort(cluster_w[v][k],cluster_w_freq[v][k],sizeof(int),
              (int(*)(const void *,const void *))compare_word);
      }
    }
  }

  xfree(tmp_count);
}

static void compute_cluster_g_ml(void)
{
  int *tmp_count;
  int k,e,u;
  int k_ml;
  FLOAT_T p_max;

  /* Allocate the temporary areas */
  tmp_count = (int *)xmalloc(sizeof(int)*num_class);

  for (u=0; u<gauss_size; u++)
    for (k=0; k<num_class; k++)
      cluster_g_freq[u][k] = 0;

  for (u=0; u<gauss_size; u++) {
    for (e=0; e<num_entry; e++) {
      k_ml = 0;
      p_max = 0.0;
      for (k=0; k<num_class; k++) {
        if (prob_mem_g[k][u][e] > p_max) {
          k_ml = k;
          p_max = prob_mem_g[k][u][e];
        }
      }
      cluster_g_pred[u][e] = k_ml;
      cluster_g_freq[u][k_ml]++;  /* Count up att-vals in class k_ml */
    }
  }

  for (u=0; u<gauss_size; u++) {
    for (k=0; k<num_class; k++)
      tmp_count[k] = 0;

    for (e=0; e<num_entry; e++) {
      k_ml = cluster_g_pred[u][e];
      cluster_g[u][k_ml][tmp_count[k_ml]] = e;
      tmp_count[k_ml]++;
    }
  }

  xfree(tmp_count);
}

static void compute_cluster_case_ml(void)
{
  int *tmp_count;
  int k,e;
  int k_ml;
  FLOAT_T p_max;

  tmp_count = (int *)xmalloc(sizeof(int)*num_class);

  for (k=0; k<num_class; k++)
    cluster_case_freq[k] = 0;

  for (e=0; e<num_entry; e++) {
    k_ml = 0;
    p_max = 0.0;
    for (k=0; k<num_class; k++) {
      if (prob_mem_case[k][e] > p_max) {
        k_ml = k;
        p_max = prob_mem_case[k][e];
      }
    }
    cluster_case_pred[e] = k_ml;
    cluster_case_freq[k_ml]++;
  }

  for (k=0; k<num_class; k++)
    tmp_count[k] = 0;

  for (e=0; e<num_entry; e++) {
    k_ml = cluster_case_pred[e];
    cluster_case[k_ml][tmp_count[k_ml]] = e;
    tmp_count[k_ml]++;
  }

  xfree(tmp_count);
}

void cluster_main(void)
{
  FILE *fp_cluster;

  if (mode_show_attr_cluster)
    print_msg("Clustering cases and attribute values...");
  else
    print_msg("Clustering cases...");
  fflush(fp_msg);

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_cluster = current_sec();
  stime_lap_cluster = elapsed_sec();
#endif

  alloc_cluster();

  if (mode_nominal) {
    switch (cl_type) {
    case CL_MP:
      compute_cluster_w_ml();
      break;
    default:
      die("Internal error: Unsupported clustering method\n");
    }
  }

  if (mode_gauss) {
    switch (cl_type) {
    case CL_MP:
      compute_cluster_g_ml();
      break;
    default:
      die("Internal error: Unsupported clustering method\n");
    }
  }

  switch (cl_type) {
  case CL_MP:
    compute_cluster_case_ml();
    break;
  default:
    die("Internal error: Unsupported clustering method\n");
  }

  print_cluster_freq(fp_log);

  fp_cluster = fopen(fname_cluster,"wb");
  if (fp_cluster == NULL) die("Couldn't open %s\n",fname_cluster);
  print_cluster(fp_cluster);
  fclose(fp_cluster);

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_cluster = current_sec();
  etime_lap_cluster = elapsed_sec();
#endif

  print_msg("done\n");

  if (mode_eval) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_eval = current_sec();
    stime_lap_eval = elapsed_sec();
#endif

    print_msg("Evaluating clusters...");
    fflush(fp_msg);
    eval_main();
    print_msg("done\n");

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_eval = current_sec();
    etime_lap_eval = elapsed_sec();
#endif
  }

  release_cluster();
}
