/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void compute_print_dissim_case(FILE *fp)
{
  int k,e0,e1;
  FLOAT_T d,d1=0.0;

  print_dissim_case_header(fp);
  for (e0=0; e0<num_entry; e0++) {
    for (e1=0; e1<num_entry; e1++) {
      d = 0.0;
      for (k=0; k<num_class; k++) {
        if (prob_mem_case[k][e0] > 0.0 && prob_mem_case[k][e1] > 0.0) {
          d1 =
            prob_mem_case[k][e0] *
            (LOG(prob_mem_case[k][e0]) - LOG(prob_mem_case[k][e1]));
        }
        else if (prob_mem_case[k][e0] > 0.0) { /* prob_mem_case[k][e1] == 0 */
          die("Couldn't compute KL div. since P(Class=%d|case-ID=%d) = 0"
              " (smoothing may be required)\n", k, e1);
        }
        else {
          d1 = 0.0;
        }
        d += d1;
      }
      print_dissim_case(fp,e0,e1,d);
    }
  }

  fflush(fp);
}

static void compute_print_dissim_w(FILE *fp)
{
  int k,v,v_d,i,j,i0,j0;
  FLOAT_T d;
  FLOAT_T d1=0.0;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    print_dissim_w_header(fp,v);

    v_d = attr_info[v].local_id;
    for (i=0; i<num_word[v_d]; i++) {
      i0 = print_order[v_d][i];
      for (j=0; j<num_word[v_d]; j++) {
        j0 = print_order[v_d][j];
        d = 0.0;
        for (k=0; k<num_class; k++) {
          if (prob_mem_w[k][v_d][i0] > 0.0 && prob_mem_w[k][v_d][j0] > 0.0) {
            d1 =
              prob_mem_w[k][v_d][i0] *
              (LOG(prob_mem_w[k][v_d][i0]) - LOG(prob_mem_w[k][v_d][j0]));
          }
          else if (prob_mem_w[k][v_d][i0] > 0.0) {
            die("Couldn't compute KL div. since P(Class=%d|%s) = 0"
                " (smoothing may be required)\n", k, word_list[v_d][j0]);
          }
          else {
            d1 = 0.0;
          }
          d += d1;
        }
        print_dissim_w(fp,v,v_d,i0,j0,d);
      }
    }
  }

  fflush(fp);
}

static void compute_print_dissim_g(FILE *fp)
{
  int k,u,e0,e1;
  FLOAT_T d,d1=0.0;

  for (u=0; u<gauss_size; u++) {
    print_dissim_g_header(fp,u);
    for (e0=0; e0<num_entry; e0++) {
      for (e1=0; e1<num_entry; e1++) {
        d = 0.0;
        for (k=0; k<num_class; k++) {
          if (prob_mem_g[k][u][e0] > 0.0 && prob_mem_g[k][u][e1] > 0.0)
            d1 =
              prob_mem_g[k][u][e0] *
              (LOG(prob_mem_g[k][u][e0]) - LOG(prob_mem_g[k][u][e1]));
          else if (prob_mem_g[k][u][e0] > 0.0)
            die("Couldn't compute KL div. since P(Class=%d|Gauss%d) = 0 for %s"
                " (smoothing should be required)\n",k,u,get_case_id(e1));
          else
            d1 = 0.0;

          d += d1;
        }
        print_dissim_g(fp,u,e0,e1,d);
      }
    }
  }

  fflush(fp);
}

void dissim_main(void)
{
  FILE *fp_dist;

  print_msg("Computing dissimilarities between attribute values...");
  fflush(fp_msg);

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_dissim = current_sec();
  stime_lap_dissim = elapsed_sec();
#endif
  fp_dist = fopen(fname_dissim,"wb"); /* binary mode */
  if (fp_dist == NULL) die("Couldn't open %s\n", fname_dissim);

  compute_print_dissim_case(fp_dist);
  if (mode_nominal) compute_print_dissim_w(fp_dist);
  if (mode_gauss) compute_print_dissim_g(fp_dist);

  fclose(fp_dist);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_dissim = current_sec();
  etime_lap_dissim = elapsed_sec();
#endif

  print_msg("done\n");
}
