/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

void determine_format(char *fname_tsv, char *fname_csv)
{
  FILE *fp;

  fp = fopen(fname_tsv,"rb");
  if (fp != NULL) {
    format_in = FMT_TSV;
    fclose(fp);
  }

  fp = fopen(fname_csv,"rb");
  if (fp != NULL) {
    format_in = FMT_CSV;
    fclose(fp);
  }

  if (format_out != FMT_CSV && format_out != FMT_TSV) format_out = format_in;

  if (format_in != FMT_CSV && format_in != FMT_TSV) {
    fprintf(stderr,
            "Couldn't find the input file (neither %s nor %s)\n",
            fname_tsv, fname_csv);
    exit(1);
  }
}

void initialize_buffer(void)
{
  buf_crlf  = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_base  = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_sys   = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_delim = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_cout  = (char *)malloc(sizeof(char)*MAX_OUT_LEN);

  if (buf_crlf == NULL || buf_base == NULL || buf_sys == NULL ||
      buf_delim == NULL || buf_cout == NULL) {
    fprintf(stderr,"Couldn't allocate memory for buffers\n");
    exit(1);
  }

  memset(buf_crlf, 0,MAX_OUT_LEN);
  memset(buf_base, 0,MAX_OUT_LEN);
  memset(buf_sys,  0,MAX_OUT_LEN);
  memset(buf_delim,0,MAX_OUT_LEN);
  memset(buf_cout, 0,MAX_OUT_LEN);
}

void release_buffer()
{
  xfree(buf_crlf);
  xfree(buf_base);
  xfree(buf_sys);
  xfree(buf_delim);
  xfree(buf_cout);
}

/* open the log files -- hereafter we can use die() */
void open_log(void)
{
  fp_log = fopen(fname_log,"wb");
  if (fp_log == NULL) {
    fprintf(stderr,"Couldn't open %s\n", fname_log);
    exit(1);
  }

  if (mode_msg) {
    fp_msg = fopen(fname_msg,"wb");
    if (fp_msg == NULL) {
      fprintf(stderr,"Couldn't open %s\n", fname_msg);
      exit(1);
    }
  }
  else {
    fp_msg = stderr;
  }
}

#ifdef USE_ICONV
void initialize_iconv_sys(void)
{
#ifdef CHARCODE
  cd_sys_in = iconv_open(CHARCODE_INTERNAL,CHARCODE);
  if (cd_sys_in == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (in: %s)"
            " -- run 'iconv -l' for available character codes\n",
            CHARCODE);

  cd_sys_out = iconv_open(CHARCODE,CHARCODE_INTERNAL);
  if (cd_sys_out == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (out: %s)"
            " -- run 'iconv -l' for available character codes\n",
            CHARCODE);
#endif
}
#endif

#ifdef USE_ICONV
void initialize_iconv()
{
#ifdef CHARCODE
  cd_sys_in = iconv_open(CHARCODE_INTERNAL,CHARCODE);
  if (cd_sys_in == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (in: %s)"
            " -- run 'iconv -l' for available character codes\n",
            CHARCODE);

  cd_sys_out = iconv_open(CHARCODE,CHARCODE_INTERNAL);
  if (cd_sys_out == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (out: %s)"
            " -- run 'iconv -l' for available character codes\n",
            CHARCODE);
#endif

  cd_in = iconv_open(CHARCODE_INTERNAL,charcode_in);
  if (cd_in == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (in: %s)"
            " -- run 'iconv -l' for available character codes\n",
            charcode_in);

  cd_out = iconv_open(charcode_out,CHARCODE_INTERNAL);
  if (cd_out == (iconv_t)-1)
    fprintf(stderr,
            "iconv_open failed (out: %s)"
            " -- run 'iconv -l' for available character codes\n",
            charcode_out);
}
#endif

/* parse the specification for the number of classes */
void parse_num_class(char *s)
{
  char *endptr;
  int i;
  char ks0[10], ks1[10], ks2[10];
  char *s0,*k0,*k1,*k2;
  int state;

  if (*s == '\0')
    die("Unexpected number of classes -- %s\n",s);

  i = strtol(s,&endptr,10);
  /* we have just an integer -- now consider this as the number of classes */
  if (*endptr == '\0') {
    num_class = i;
    return;
  }

  /* try to parse the string in the form of Kmin:Kmax:Kstep */
  state = 0;
  k0 = ks0;
  k1 = ks1;
  k2 = ks2;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ':' || *s0 == ',' || *s0 == '-') {
      state++;
    }
    else {
      if (state == 0) {
        *k0 = *s0;
        k0++;
      }
      else if (state == 1) {
        *k1 = *s0;
        k1++;
      }
      else if (state == 2) {
        *k2 = *s0;
        k2++;
      }
    }
    s0++;
  }

  if (state != 2) die("Illegal value for %s option\n",OPTSTR(O_N_CLASS));

  *k0 = *k1 = *k2 = '\0';

  kmin = parse_int(ks0);
  kmax = parse_int(ks1);
  kstep = parse_int(ks2);
  mode_search_num_class = 1;
}

#ifdef VNBC
static void check_smooth(FLOAT_T f, char *s)
{
  if (f <= 0) die("Unexpected value -- %s\n",s);
}
#else
static void check_smooth(FLOAT_T f, char *s)
{
  if (f < 0)  die("Unexpected value -- %s\n",s);
}
#endif

FLOAT_T parse_smooth_c(char *s)
{
  FLOAT_T f = 0.0;

  if (strcmp(s,SMOOTH_AUTO) == 0) {
    f = -1.0;
    mode_smooth_auto = 1;
    mode_smooth_auto_c = MODE_SMOOTH_AUTO;
  }
  else {
    f = parse_float(s);
    check_smooth(f,s);

    mode_smooth_auto_c = MODE_SMOOTH_NONE;
  }

  return f;
}

/* parse the specification for attribute pseudo counts 'u1,u2,u3,..' */
int parse_smooth_w(char *s)
{
  char buf[MAX_FLOAT_LEN];  /* MAX_FLOAT_LEN is assumed to be positive */
  char *s0;
  int n,k,num_smooth_w;

  /* We first count up the number of commas: */
  n = 0;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ',') n++;
    s0++;
  }
  num_smooth_w = n + 1;

  if (smooth_w != NULL) xfree(smooth_w);
  smooth_w = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_smooth_w);

  if (mode_smooth_auto_w != NULL) xfree(mode_smooth_auto_w);
  mode_smooth_auto_w =
    (enum smooth_auto_type *)
    xmalloc(sizeof(enum smooth_auto_type)*num_smooth_w);

  n = 0;
  k = 0;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ',') {
      buf[k] = '\0';
      if (strcmp(buf,SMOOTH_AUTO) == 0) {
        smooth_w[n] = -1.0;
        mode_smooth_auto = 1;
        mode_smooth_auto_w[n] = MODE_SMOOTH_AUTO;
      }
      else {
        smooth_w[n] = parse_float(buf);
        check_smooth(smooth_w[n],buf);
        mode_smooth_auto_w[n] = MODE_SMOOTH_NONE;
      }
      n++;
      k = 0;
    }
    else {
      buf[k] = *s0;
      k++;
    }

    if (k >= MAX_FLOAT_LEN && k > strlen(SMOOTH_AUTO))
      die("Too long input for a floating-point number: %s\n",s);

    s0++;
  }
  buf[k] = '\0';
  if (strcmp(buf,SMOOTH_AUTO) == 0) {
    smooth_w[n] = -1.0;
    mode_smooth_auto_w[n] = MODE_SMOOTH_AUTO;
  }
  else {
    smooth_w[n] = parse_float(buf);
    check_smooth(smooth_w[n],buf);
    mode_smooth_auto_w[n] = MODE_SMOOTH_NONE;
  }

  return num_smooth_w;
}

void parse_smooth_s(char *s)
{
  char *s0,*s1;

  s0 = strchr(s,':');

  if (s0 == NULL) {
    if (strcmp(s,SMOOTH_VAR_STR_SAMPLE) == 0) {
      smooth_s = 1.0;
      mode_smooth_s = SMOOTH_VAR_SAMPLE;
    }
    else if (strcmp(s,SMOOTH_VAR_STR_CONST) == 0) {
      smooth_s = 1.0;
      mode_smooth_s = SMOOTH_VAR_CONST;
    }
    else {
      smooth_s = parse_float(s);
      if (smooth_s < 0)
        die("Illegal value for %s option\n",OPTSTR(O_SMOOTH_VAR));
      mode_smooth_s = SMOOTH_VAR_SAMPLE;
    }
  }
  else {
    s1 = dup_str(s);
    s1[(s0 - s)] = '\0';
    if (strcmp(s1,SMOOTH_VAR_STR_SAMPLE) == 0) {
      mode_smooth_s = SMOOTH_VAR_SAMPLE;
    }
    else if (strcmp(s1,SMOOTH_VAR_STR_CONST) == 0) {
      mode_smooth_s = SMOOTH_VAR_CONST;
    }
    else {
      die("Illegal value for %s option\n",OPTSTR(O_SMOOTH_VAR));
    }
    smooth_s = parse_float(s0 + 1);
    if (smooth_s < 0)
      die("Illegal value for %s option\n",OPTSTR(O_SMOOTH_VAR));
    xfree(s1);
  }
}

FLOAT_T parse_smooth_tau(char *s)
{
  FLOAT_T f = 0.0;

  if (strcmp(s,SMOOTH_AUTO) == 0) {
    f = -1.0;
    mode_smooth_auto = 1;
    mode_smooth_auto_tau = MODE_SMOOTH_AUTO;
  }
  else {
    f = parse_float(s);
    if (f < 0) die("Illegal value for %s option\n",OPTSTR(O_TAU));

    mode_smooth_auto_tau = MODE_SMOOTH_NONE;
  }

  return f;
}

void build_smooth_nominal(int num_smooth_w)
{
  int v;
  FLOAT_T tmp_smooth_w;
  enum smooth_auto_type tmp_mode_smooth_auto_w;

  if (tuple_size_d != num_smooth_w) {
    if (num_smooth_w == 0) {  /* if -w option is not given */
      smooth_w = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*tuple_size_d);

      mode_smooth_auto_w =
        (enum smooth_auto_type *)
        xmalloc(sizeof(enum smooth_auto_type)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
        smooth_w[v] = smooth_default;
        mode_smooth_auto_w[v] = MODE_SMOOTH_NONE;
      }
    }
    else if (num_smooth_w == 1) {
      /* Reallocate smooth_w[] when uniform smoothing constants specified: */
      tmp_smooth_w = smooth_w[0];
      tmp_mode_smooth_auto_w = mode_smooth_auto_w[0];

      xfree(smooth_w);
      xfree(mode_smooth_auto_w);

      smooth_w = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*tuple_size_d);
      mode_smooth_auto_w =
        (enum smooth_auto_type *)
        xmalloc(sizeof(enum smooth_auto_type)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
        smooth_w[v] = tmp_smooth_w;
        mode_smooth_auto_w[v] = tmp_mode_smooth_auto_w;
      }
    }
    else
      die("Too many or too few smoothing constants for %s\n",
          OPTSTR(O_SMOOTH_A));
  }
}

FLOAT_T get_tau(int k)
{
  return (mode_read_hparam) ? hparam_tau[k] : tau;
}

#ifdef VNBC
FLOAT_T get_alpha_w(int n)
{
  return (alpha_w > 0) ? alpha_w : (n + ALPHA_W_NOISE);
}
#else
FLOAT_T get_alpha_w(int k, int u, int n)
{
  return
    (mode_read_hparam) ?
    hparam_alpha_w[k][u] : ((alpha_w > 0) ? alpha_w : (n + ALPHA_W_NOISE));
}
#endif

/* used only for nbc */
FLOAT_T get_hparam_c(int k)
{
  return (mode_read_hparam) ? hparam_c[k] : (smooth_c + 1.0);
}

/* used only for nbc */
FLOAT_T get_hparam_w(int k, int v, int i)
{
  return (mode_read_hparam) ? hparam_w[k][v][i] : (smooth_w[v] + 1.0);
}

void check_read_hparam(void)
{
  int k,v,i,u;

  mode_smooth = 0;
  mode_smooth_strict = 1;

  for (k=0; k<num_class; k++) {
    mode_smooth |= (hparam_c[k] > 1.0) ? 1 : 0;
    mode_smooth_strict &= (hparam_c[k] > 1.0) ? 1 : 0;
  }

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        for (i=0; i<num_word[v]; i++) {
          mode_smooth |= (hparam_w[k][v][i] > 1.0) ? 1 : 0;
          mode_smooth_strict &= (hparam_w[k][v][i] > 1.0) ? 1 : 0;
        }
      }
    }

  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      if (hparam_tau[k] <= 0.0)
        die("Illegal tau (%.Nf,class=%d) read from %s\n",
            hparam_tau[k],k,fname_hparam);

      for (u=0; u<gauss_size; u++) {
        if (hparam_alpha_w[k][u] < gauss_attr_size[u])
          fprintf(stderr,
                  conv_msg("Illegal alpha (%.Nf, class=%d, dist=%d)"
                           " read from %s\n"),
                  hparam_alpha_w[k][u],k,u,fname_hparam);
      }
    }

    mode_smooth_gauss = 1;

    mode_smooth |= mode_smooth_gauss;
    mode_smooth_strict &= mode_smooth_gauss;
  }

  if (!mode_smooth_strict) {
    fprintf(stderr,
            "Pseudo counts read from %s should be all positive\n",
            fname_hparam);
    exit(1);
  }
}

void parse_extreme(char *s)
{
  FLOAT_T extreme;

  extreme = parse_float(s);

  if (extreme < 0 || extreme > 1)
    die("Illegal value for %s option\n",OPTSTR(O_EXTREME));

  q_regular_min = extreme;
  q_regular_max = 1.0 - extreme;
}

void parse_itemp(char *s)
{
  char buf[MAX_FLOAT_LEN];
  char *s0;
  int n,k,num_field;

  n = 0;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ',' || *s0 == ':') n++;
    s0++;
  }
  num_field = n + 1;

  if (num_field != 2)
    die("Illegal specification of the inverse temperature -- %s\n",s);

  s0 = s;
  k = 0;
  while (*s0 != '\0') {
    if (*s0 == ',' || *s0 == ':') {
      buf[k] = '\0';
      if (strlen(buf) > 0) itemp_init = parse_float(buf);
      k = 0;
    }
    else {
      buf[k] = *s0;
      k++;
    }
    s0++;
  }
  buf[k] = '\0';
  if (strlen(buf) > 0) itemp_rate = parse_float(buf);

  if (itemp_init <= 0 || itemp_init > 1)
    die(conv_msg("Illegal value for the initial value of inverse temperature"
                 " -- %.Nf\n"),
        itemp_init);

  if (!(itemp_rate > 1))
    die(conv_msg("Illegal value for the increasing rate of inverse temperature"
                 " -- %.Nf\n"),
        itemp_rate);
}

void parse_rank(char *s)
{
  char *endptr_i,*endptr_d;
  int i;
  FLOAT_T d;

  if (*s == '\0') die("Null string specified for a floating-point number\n");

  i = strtol(s, &endptr_i, 10);
  d = STRTOD(s, &endptr_d);

  /* parsed as an integer */
  if (*endptr_i == '\0' && i >= 0) {
    num_rank = i;
  }
  /* parsed as a floating point number */
  else if (*endptr_d == '\0' && d < 1 && d > 0) {
    mode_rank_th = 1;
    rank_th = d;
  }
  else
    die("Illegal specification for ranking -- %s\n",s);
}

/* parse a string representation of integer */
int parse_int(char *s)
{
  char *endptr;
  int i;

  if (*s == '\0') die("Null string specified for an integer\n");

  i = strtol(s, &endptr, 10);
  if (*endptr != '\0')
    die("Couldn't parse as an integer -- %s\n",s);

  return i;
}

/* parse a string representation of unsigned integer */
unsigned int parse_uint(char *s)
{
  char *endptr;
  unsigned int i;

  if (*s == '\0') die("Null string specified for an unsigned integer\n");

  i = strtoul(s, &endptr, 10);
  if (*endptr != '\0')
    die("Couldn't parse as an unsigned integer -- %s\n",s);

  return i;
}

/* parse a string representation of floating-point numbers */
FLOAT_T parse_float(char *s)
{
  char *endptr;
  FLOAT_T d;

  if (*s == '\0') die("Null string specified for a floating-point number\n");

  d = STRTOD(s, &endptr);
  if (*endptr != '\0')
    die("Couldn't parse as floating-point number -- %s\n",s);

  return d;
}

/* set the random seed */
void initialize_rand()
{
  random_seed(rand_seed);
}

/* set the pseudo counts based on the current number of classes */
void set_smooth_auto(void)
{
  int v;

  if (mode_nominal) {
    if (mode_smooth_auto_c == MODE_SMOOTH_AUTO)
      smooth_c = 1.0 / num_class;

    for (v=0; v<tuple_size_d; v++) {
      if (mode_smooth_auto_w[v] == MODE_SMOOTH_AUTO)
        smooth_w[v] = 1.0 / num_word[v];
    }
  }

  if (mode_gauss) {
    if (mode_smooth_auto_c == MODE_SMOOTH_AUTO)
      smooth_c = 1.0 / num_class;

    if (mode_smooth_auto_tau == MODE_SMOOTH_AUTO)
      tau = 1.0 / num_class;
  }

  if (mode_smooth_auto) print_smooth(fp_log, num_class);

  mode_smooth = (smooth_c > 0) ? 1 : 0;
  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      mode_smooth |= (smooth_w[v] > 0) ? 1 : 0;
  }
  if (mode_gauss) {
    mode_smooth |= (mode_smooth_gauss) ? 1 : 0;
  }

  mode_smooth_strict = (smooth_c > 0) ? 1 : 0;
  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      mode_smooth_strict &= (smooth_w[v] > 0) ? 1 : 0;
    for (v=0; v<tuple_size_d; v++) {
    }
  }
  if (mode_gauss) {
    mode_smooth_strict &= (mode_smooth_gauss) ? 1 : 0;
  }

  if (!mode_smooth_strict && ic_type == AUTOCLASS)
    die("Couldn't use the AUTOCLASS mode"
        " unless all smoothing constant > 0\n");
}

static void check_ignore_single(int t, char *s)
{
  if (t < 0 || t >= tuple_size)
    die("Out-of-range specification of an ignored attribute -- %s\n",s);
  if (is_attr_id(t))
    die("Confliction between %s and %s options\n",
        OPTSTR(O_ID),OPTSTR(O_IGNORE));
  if (is_attr_eval(t))
    die("Confliction between %s and %s options\n",
        OPTSTR(O_EVAL),OPTSTR(O_IGNORE));
}

static void check_ignore_range(int min, int max, char *s)
{
  int t;

  if (min > max)
    die("Illegal specification of ignored attributes -- %s\n",s);

  for (t=min; t<=max; t++)
    check_ignore_single(t,s);
}

static void add_attr_ignore0(char *s, int *attr_occ)
{
  char *q,*s0;
  int num_hyphen,pos_hyphen;
  int len,min,max,t;

  len = strlen(s);
  s0 = dup_str(s);

  num_hyphen = 0;
  pos_hyphen = -1;

  q = s0;
  while (*q != '\0') {
    if (*q == RANGE_IGNORE) {
      num_hyphen++;
      pos_hyphen = (q - s0);
    }
    q++;
  }

  if (num_hyphen == 0) {
    min = parse_int(s);
    check_ignore_single(min,s);
    attr_occ[min]++;
  }
  else if (num_hyphen == 1) {
    if (pos_hyphen == 0 && pos_hyphen == len - 1) {
      min = 0;
      max = tuple_size - 1;
    }
    else if (pos_hyphen == 0) {
      min = 0;
      max = parse_int(s0 + 1);
    }
    else if (pos_hyphen == len - 1) {
      s0[pos_hyphen] = '\0';
      min = parse_int(s0);
      max = tuple_size - 1;
    }
    else {
      s0[pos_hyphen] = '\0';
      min = parse_int(s0);
      max = parse_int(s0 + pos_hyphen + 1);
    }
    check_ignore_range(min,max,s);

    for (t=min; t<=max; t++)
      attr_occ[t]++;
  }
  else
    die("Illegal specification for ignored attribute(s) -- %s\n",s);

  xfree(s0);
}

void add_attr_ignore(char *s)
{
  char *s0,*q,*r;
  int *attr_occ;
  int v;

  if (attr_ignore == NULL) {
    attr_ignore = (int *)xmalloc(sizeof(int)*tuple_size);
    for (v=0; v<tuple_size; v++)
      attr_ignore[v] = 0;
  }

  attr_occ = (int *)xmalloc(sizeof(int)*tuple_size);
  for (v=0; v<tuple_size; v++)
    attr_occ[v] = 0;

  s0 = dup_str(s);

  q = s0;
  r = s0;
  while (1) {
    if (*q == ',') {
      *q = '\0';
      add_attr_ignore0(r,attr_occ);
      r = q + 1;
    }
    else if (*q == '\0') {
      add_attr_ignore0(r,attr_occ);
      break;
    }
    q++;
  }

  for (v=0; v<tuple_size; v++)
    if (attr_occ[v] > 0) attr_ignore[v] = 1;

  xfree(attr_occ);
  xfree(s0);
}

static void check_gauss_single(int t, char *s)
{
  if (t < 0 || t >= tuple_size)
    die("Out-of-range specification of a continuous attribute -- %s\n",s);
  if (is_attr_id(t))
    die("Confliction between %s and %s options\n",
        OPTSTR(O_GAUSS),OPTSTR(O_ID));
  if (is_attr_eval(t))
    die("Confliction between %s and %s options\n",
        OPTSTR(O_GAUSS),OPTSTR(O_EVAL));
  if (is_attr_ignore(t))
    die("Confliction between %s and %s options\n",
        OPTSTR(O_GAUSS),OPTSTR(O_IGNORE));
}

static void check_gauss_range(int min, int max, char *s)
{
  int t;

  if (min > max)
    die("Illegal specification of continuous attributes -- %s\n",s);

  for (t=min; t<=max; t++)
    check_gauss_single(t,s);
}

static void add_gauss0(char *s)
{
  GAUSS_PTR g,g0;
  int min,max,num_sep,pos_sep,len,t;
  char *q;
  char *s0;

  if (gauss_list == NULL) {
    gauss_list = (GAUSS_PTR)xmalloc(sizeof(struct GAUSS));  /* dummy cell */

    gauss_list->str = NULL;
    gauss_list->size = 0;
    gauss_list->attr = NULL;
    gauss_list->next = NULL;
  }

  g = gauss_list;
  while (g->next != NULL) {
    g = g->next;
  }

  num_sep = 0;
  pos_sep = -1;
  q = s;
  while (*q != '\0') {
    if (*q == RANGE_GAUSS_UV) {
      num_sep++;
      pos_sep = (q - s);
    }
    q++;
  }

  if (num_sep == 1) {
    len = strlen(s);
    s0 = dup_str(s);

    if (pos_sep == 0 && pos_sep == len - 1) {
      min = 0;
      max = tuple_size - 1;
    }
    else if (pos_sep == 0) {
      min = 0;
      max = parse_int(s0 + 1);
    }
    else if (pos_sep == len - 1) {
      s0[pos_sep] = '\0';
      min = parse_int(s0);
      max = tuple_size - 1;
    }
    else {
      s0[pos_sep] = '\0';
      min = parse_int(s0);
      max = parse_int(s0 + pos_sep + 1);
    }
    check_gauss_range(min,max,s);

    for (t=min; t<=max; t++) {
      g0 = (GAUSS_PTR)xmalloc(sizeof(struct GAUSS));
      g0->str = int2strdup(t);
      g0->size = 0;
      g0->attr = NULL;
      g0->next = NULL;

      g->next = g0;
      g = g->next;
    }

    xfree(s0);
  }
  else if (num_sep > 1) {
    die("Illegal specification for continuous attribute(s) -- %s\n",s);
  }
  else { /* num_sep == 0 */
    g0 = (GAUSS_PTR)xmalloc(sizeof(struct GAUSS));
    g0->str = dup_str(s);

    g0->size = 0;
    g0->attr = NULL;
    g0->next = NULL;

    g->next = g0;
  }
}

/* add a set of attributes which follows (multivariate) Gaussian
 *  -u 2-5,8#10,13-15#20=22 -u 35-38
 * (though we only add the original string reprentation)
 */
void add_gauss(char *s)
{
  char *s0,*q,*r;

  s0 = dup_str(s);

  q = s0;
  r = s0;
  while (1) {
    if (*q == SEP_GAUSS) {
      *q = '\0';
      add_gauss0(r);
      r = q + 1;
    }
    else if (*q == '\0') {
      add_gauss0(r);
      break;
    }
    q++;
  }

  xfree(s0);
}

static void clean_gauss(int clean_attr)
{
  GAUSS_PTR g,g0;

  if (gauss_list == NULL) return;

  g = gauss_list->next;
  while (g != NULL) {
    if (g->str != NULL) xfree(g->str);
    if (clean_attr && g->attr != NULL) xfree(g->attr);
    g0 = g;
    g = g->next;
    xfree(g0);
  }

  xfree(gauss_list);
}

#if 0
static void print_gauss_attr(void)
{
  int a,u;

  for (u=0; u<gauss_size; u++) {
    fprintf(stderr,">> gauss[%d]:",u);
    for (a=0; a<gauss_attr_size[u]; a++) {
      fprintf(stderr," %d",gauss_attr[u][a]);
    }
    fprintf(stderr,"\n");
  }
}

static void print_attr_info(void)
{
  int v;

  for (v=0; v<tuple_size; v++) {
    if (is_attr_cont(v)) {
      fprintf(stderr,"> attr_info[%d] (cont) gauss=%d, local=%d\n",
              v,attr_info[v].gauss_id,attr_info[v].local_id);
    }
    else {
      fprintf(stderr,"> attr_info[%d] (disc) local=%d\n",
              v,attr_info[v].local_id);
    }
  }
}
#endif

/* attr_occ[] indicate the total numbers of occurrences */
static void parse_gauss1(char *s, int *attr_occ)
{
  int len;
  char *s0,*q;
  int num_hyphen,pos_hyphen;
  int t,min,max;

  len = strlen(s);
  s0 = dup_str(s);

  num_hyphen = 0;
  pos_hyphen = -1;

  q = s0;
  while (*q != '\0') {
    if (*q == RANGE_GAUSS_MV) {
      num_hyphen++;
      pos_hyphen = (q - s0);
    }
    q++;
  }

  if (num_hyphen == 0) {
    min = parse_int(s0);
    check_gauss_single(min,s);
    attr_occ[min]++;
  }
  else if (num_hyphen == 1) {
    if (pos_hyphen == 0 && pos_hyphen == len - 1) {
      min = 0;
      max = tuple_size - 1;
    }
    else if (pos_hyphen == 0) {
      min = 0;
      max = parse_int(s0 + 1);
    }
    else if (pos_hyphen == len - 1) {
      s0[pos_hyphen] = '\0';
      min = parse_int(s0);
      max = tuple_size - 1;
    }
    else {
      s0[pos_hyphen] = '\0';
      min = parse_int(s0);
      max = parse_int(s0 + pos_hyphen + 1);
    }
    check_gauss_range(min,max,s);

    for (t=min; t<=max; t++)
      attr_occ[t]++;
  }
  else
    die("Illegal specification for continuous attribute(s) -- %s\n",s);

  xfree(s0);
}

static void parse_gauss0(char *s, int *attr_occ, int *attr_occ0)
{
  char *s0,*q,*r;
  int v;

  s0 = dup_str(s);

  q = s0;
  r = s0;
  while (1) {
    if (*q == ',') {
      *q = '\0';
      parse_gauss1(r,attr_occ);
      r = q + 1;
    }
    else if (*q == '\0') {
      parse_gauss1(r,attr_occ);
      break;
    }
    q++;
  }

  for (v=0; v<tuple_size; v++) {
    if (attr_occ[v] > 0) attr_occ0[v]++;
  }

  xfree(s0);
}

#if 0
static void print_gauss_list(void)
{
  GAUSS_PTR g;
  int u;

  if (gauss_list == NULL) return;

  u = 0;
  g = gauss_list->next;
  while (g != NULL) {
    fprintf(stderr,"> gauss[%d]:",u);
    fprintf(stderr," %s\n",g->str);
    g = g->next;
    u++;
  }
}
#endif

void parse_gauss(void)
{
  GAUSS_PTR g;
  int *attr_occ,*attr_occ0;
  char *s;
  int v,a,size;

  if (gauss_list == NULL) return;

  attr_occ = (int *)xmalloc(sizeof(int)*tuple_size);
  attr_occ0 = (int *)xmalloc(sizeof(int)*tuple_size);

  for (v=0; v<tuple_size; v++)
    attr_occ0[v] = 0;

  g = gauss_list->next;
  while (g != NULL) {
    s = g->str;

    for (v=0; v<tuple_size; v++)
      attr_occ[v] = 0;

    parse_gauss0(s,attr_occ,attr_occ0);

    size = 0;
    for (v=0; v<tuple_size; v++) {
      if (attr_occ[v] > 0) size++;
    }

    g->size = size;
    g->attr = (int *)xmalloc(sizeof(int)*size);

    a = 0;
    for (v=0; v<tuple_size; v++) {
      if (attr_occ[v] > 0) {
        g->attr[a] = v;
        a++;
      }
    }

    g = g->next;
  }

  for (v=0; v<tuple_size; v++) {
    if (attr_occ0[v] > 1)
      die("Overlapping continuous attrubute: %d\n",v);
  }

  xfree(attr_occ);
  xfree(attr_occ0);
}

void check_tuple_size(void)
{
  int tuple_size0;
  int v;

  tuple_size0 = tuple_size;
  if (mode_id) tuple_size0--;
  if (mode_eval) tuple_size0--;
  if (mode_ignore)
    for (v=0; v<tuple_size; v++) {
      if (is_attr_ignore(v)) tuple_size0--;
    }

  if (tuple_size0 <= 0)
    die("No descriptive attributes\n");
}

void build_attr_info(void)
{
  GAUSS_PTR g;
  int u,v,a,l;

  parse_gauss();

  gauss_size = 0;
  tuple_size_c = 0;

  if (gauss_list != NULL) {
    g = gauss_list->next;
    while (g != NULL) {
      gauss_size++;
      tuple_size_c += g->size;
      g = g->next;
    }
  }

  tuple_size_d = tuple_size;
  tuple_size_d -= tuple_size_c;
  if (mode_eval) tuple_size_d -= 1;
  if (mode_id) tuple_size_d -= 1;
  if (mode_ignore)
    for (v=0; v<tuple_size; v++)
      if (attr_ignore[v]) tuple_size_d -= 1;
  tuple_size_all = tuple_size_d + tuple_size_c;

  mode_nominal = (tuple_size_d > 0) ? 1 : 0;

  if (gauss_list != NULL) {
    gauss_attr = (int **)xmalloc(sizeof(int *)*gauss_size);
    gauss_attr_size = (int *)xmalloc(sizeof(int)*gauss_size);

    g = gauss_list->next;
    u = 0;
    while (g != NULL) {
      gauss_attr_size[u] = g->size;
      gauss_attr[u] = g->attr;  /* shallow copy */
      g = g->next;
      u++;
    }
  }

  attr_info = (ATTR_INFO *)xmalloc(sizeof(ATTR_INFO)*tuple_size);
  for (v=0; v<tuple_size; v++) {
    attr_info[v].is_cont = 0;
  }
  for (u=0; u<gauss_size; u++) {
    for (a=0; a<gauss_attr_size[u]; a++) {
      v = gauss_attr[u][a];
      attr_info[v].is_cont = 1;
      attr_info[v].gauss_id = u;
      attr_info[v].local_id = a;
    }
  }
  l = 0;
  for (v=0; v<tuple_size; v++) {
    if (is_attr_id(v)     || is_attr_eval(v) ||
        is_attr_ignore(v) || is_attr_cont(v))
      continue;
    attr_info[v].local_id = l;
    l++;
  }

#if 0
  print_gauss_attr();
  print_attr_info();
#endif

  clean_gauss(0);
}

void initialize_hparam_gauss(void)
{
  int k,u,a,n;

  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      if (n > 1) {
        if (mode_smooth_u)
          for (a=0; a<n; a++) hparam_u[k][u][a] = smooth_u;
        else
          copy_vector(hparam_u[k][u],sample_u[u]);

        switch (mode_smooth_s) {
        case SMOOTH_VAR_SAMPLE:
          copy_matrix(hparam_s[k][u],sample_s[u]);
          hparam_sdet[k][u] = INVERT(hparam_s[k][u],hparam_p[k][u]);
          mul_const_matrix(hparam_s[k][u],smooth_s);
          div_const_matrix(hparam_p[k][u],smooth_s);
          hparam_sdet[k][u] *= POW(smooth_s,n);
          break;
        case SMOOTH_VAR_CONST:
          set_diag_matrix(hparam_s[k][u],smooth_s);
          set_diag_matrix(hparam_p[k][u],1.0/smooth_s);
          hparam_sdet[k][u] = POW(smooth_s,n);
          break;
        default:
          die("Internal error: Illegal mode_smooth_s\n");
        }
      }
      else {
        hparam_u[k][u][0] = (mode_smooth_u) ? smooth_u : sample_u[u][0];

        switch (mode_smooth_s) {
        case SMOOTH_VAR_SAMPLE:
          hparam_s[k][u][0][0] = smooth_s * sample_s[u][0][0];
          break;
        case SMOOTH_VAR_CONST:
          hparam_s[k][u][0][0] = smooth_s;
          break;
        default:
          die("Internal error: Illegal mode_smooth_s\n");
        }

        hparam_p[k][u][0][0] = 1.0/hparam_s[k][u][0][0];
        hparam_sdet[k][u] = hparam_s[k][u][0][0];
      }
    }
  }
}
