/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_var.h"

static int num_smooth_w = 0;

static void initialize_mode()
{
  memset(fname_base,0,MAX_FILENAME);

  num_class = 0;
  tuple_size = 0;
  kmin = -1;
  kmax = -1;
  kstep = -1;
  epsilon = DEFAULT_EPSILON;
  max_iterate = DEFAULT_MAX_ITERATE;
  rand_seed = DEFAULT_RAND_SEED;
  smooth_default = 0;
  smooth_c = smooth_default;
  smooth_w = NULL;
  ic_type = BIC;
  rel_m = REL_MP;
  init_type_class = NOISY_UNIFORM_C;
  init_type_word = NOISY_UNIFORM_W;
  mode_verbose = 0;
  mode_msg = 0;
  format_in = -1;
  format_out = -1;
#ifdef USE_ICONV
#ifdef CHARCODE
  strncpy(charcode_in,CHARCODE,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE,CHARCODE_LEN);
#else
  strncpy(charcode_in,CHARCODE_INTERNAL,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE_INTERNAL,CHARCODE_LEN);
#endif
#endif
  mode_em = 1;
  mode_smooth = 0;
  mode_smooth_strict = 0;
  mode_smooth_auto = 0;
  mode_smooth_auto_c = MODE_SMOOTH_NONE;
  mode_smooth_auto_w = NULL;
  mode_smooth_auto_tau = MODE_SMOOTH_NONE;
  mode_smooth_gauss = 0;
  mode_memp = 0;
  mode_cluster = 1;
  mode_dissim = 0;
  mode_as_is = 0;
  mode_eval = 0;
  attr_eval = -1;
  mode_read_hparam = 0;
  mode_search_num_class = 0;
  mode_suffix = 0;
  num_init = 1;
  max_init_iterate = 0;
  mode_set_seed = 0;
  mode_lex_order = 1;
  mode_rank = 0;
  mode_rank_hide = 0;
  mode_rank_th = 0;
  num_rank = -1;
  rank_th = -1;
  num_sm = 0;
  max_candidate = 0;
  max_jsplit = 0;
  mode_norm = 0;
  min_param_c = DEFAULT_MIN_PARAM_C;
  min_smem_improve = -1;
  min_new_param_c = DEFAULT_MIN_NEW_PARAM_C;
  mode_daem = 0;
  itemp_init = DEFAULT_ITEMP_INIT;
  itemp_rate = DEFAULT_ITEMP_RATE;
  mode_gauss = 0;
  gauss_list = NULL;
  mode_smooth_u = 0;
  mode_smooth_s = SMOOTH_VAR_SAMPLE;
  smooth_u = 0;
  smooth_s = DEFAULT_SMOOTH_S;
  tau = 0;
  alpha_w = 0;
  mode_log_scale = 0;
  mode_id = 0;
  attr_id = -1;
  mode_ignore = 0;
  attr_ignore = NULL;
  mode_hide_attr = 0;
  cl_type = CL_MP;
  prec_out = DEFAULT_PREC_OUT;
  prec_msg = DEFAULT_PREC_MSG;
  mode_uncompress = 0;
  weight_purity = 1.0;
  weight_wallance = 1.0;
  mode_show_attr_memp = 0;
  mode_show_attr_cluster = 0;
  noise_ratio = NOISE_RATIO;
  noise_ratio_gauss = NOISE_RATIO_GAUSS;
  q_regular_min = Q_REGULAR_MIN;
  q_regular_max = Q_REGULAR_MAX;
  mode_regular_stat = 0;

#ifdef USE_ICONV
#  ifdef CHARCODE
  mode_sys_out = (strcmp(CHARCODE,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#  endif
#endif
}

static void build_filename(void)
{
  int i,l;
  char tmp_fname_tsv[MAX_FILENAME_AUX];
  char tmp_fname_csv[MAX_FILENAME_AUX];
  char *fname_ary[] = { fname_freq, fname_param, fname_hparam,  fname_dissim,
                        fname_memp, fname_log,   fname_msg,     fname_cluster,
                        fname_rank, fname_eval,  tmp_fname_tsv, tmp_fname_csv
                      };
  int num_fname = (int)(sizeof(fname_ary)/sizeof(char *));

  /* get the filenames from basename, suffix (it might be better to use snprintf()) */
  for (i=0; i<num_fname; i++) {
    memset(fname_ary[i],0,MAX_FILENAME_AUX);
    strncpy(fname_ary[i], fname_base, MAX_FILENAME);
    if (mode_suffix && fname_ary[i] != fname_freq) {
      strncat(fname_ary[i], "_", 1);
      strncat(fname_ary[i], fname_suffix, MAX_FILENAME);
    }
  }

  l = MAX_FSX_LEN + MAX_FEXT_LEN + 2;

  /* Build the filename of the input */
  if (mode_em) {
    strncpy(tmp_fname_tsv, fname_freq, MAX_FILENAME_AUX);
    strncpy(tmp_fname_csv, fname_freq, MAX_FILENAME_AUX);
    strncat(tmp_fname_tsv, "_" FSX_FREQ "." FEXT_TSV, l);
    strncat(tmp_fname_csv, "_" FSX_FREQ "." FEXT_CSV, l);
  }
  else {
    strncpy(tmp_fname_tsv, fname_param, MAX_FILENAME_AUX);
    strncpy(tmp_fname_csv, fname_param, MAX_FILENAME_AUX);
    strncat(tmp_fname_tsv, "_" FSX_PARAM "." FEXT_TSV, l);
    strncat(tmp_fname_csv, "_" FSX_PARAM "." FEXT_CSV, l);
  }

  /* format_in and format_out will be fixed here */
  determine_format(tmp_fname_tsv,tmp_fname_csv);

  strncat(fname_freq,   ((format_in == FMT_CSV) ?  "_" FSX_FREQ "." FEXT_CSV  : "_" FSX_FREQ "." FEXT_TSV), l);
  strncat(fname_param,  (((mode_em && format_out == FMT_CSV) || (!mode_em && format_in == FMT_CSV)) ?
                         "_" FSX_PARAM "." FEXT_CSV : "_" FSX_PARAM ".txt"), l);
  strncat(fname_hparam, ((format_in == FMT_CSV) ? "_" FSX_HPARAM "." FEXT_CSV : "_" FSX_HPARAM "." FEXT_TSV), l);
  strncat(fname_dissim, ((format_out == FMT_CSV) ? "_" FSX_DIST  "." FEXT_CSV : "_" FSX_DIST   "." FEXT_TSV), l);
  strncat(fname_memp,   ((format_out == FMT_CSV) ? "_" FSX_MEMP  "." FEXT_CSV : "_" FSX_MEMP   "." FEXT_TSV), l);
  strncat(fname_log,    ((format_out == FMT_CSV) ? "_" FSX_LOG   "." FEXT_CSV : "_" FSX_LOG    "." FEXT_TSV), l);
  strncat(fname_msg,    "_" FSX_MSG "." FEXT_TSV, l);
  strncat(fname_cluster, ((format_out == FMT_CSV) ? "_" FSX_CLUSTER "." FEXT_CSV : "_" FSX_CLUSTER "." FEXT_TSV), l);
  strncat(fname_rank,    ((format_out == FMT_CSV) ? "_" FSX_RANK "." FEXT_CSV : "_" FSX_RANK "." FEXT_TSV), l);
  strncat(fname_eval,    ((format_out == FMT_CSV) ? "_" FSX_EVAL "." FEXT_CSV : "_" FSX_EVAL "." FEXT_TSV), l);
}

static void parse_args(int argc, char *argv[])
{
  int c;

  /* Firstly check -h and the options that relate to log files
   * (filenames etc.)
   */
  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case GETOPT_ERR:
      fprintf(stderr,"Unknown/illegal option specification: %s (Use -h for usage)\n",optstr);
      exit(1);
    case O_TARGET:
      copy_optarg(fname_base,MAX_FILENAME);
      break;
    case O_VERSION:
      print_version(stdout);
      exit(0);
    case O_HELP:
      print_usage(stdout);
      exit(0);
    case O_XHELP:
      print_usage_full(stdout);
      exit(0);
    default:
      break;
    }
  }

  if (fname_base[0] == '\0') {
    fprintf(stderr,
            "Missing or illegal %s option (Use -h for usage)\n",
            OPTSTR(O_TARGET));
    exit(1);
  }

  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case GETOPT_ERR:
      fprintf(stderr,
              "Unknown/illegal option specification: %s (Use -h for usage)\n",
              optstr);
      exit(1);
    case O_INPUT:
      tolower_str(xoptarg_c);
      if (strcmp(xoptarg_c,FEXT_CSV) == 0) {
        format_in = FMT_CSV;
      }
      else if (strcmp(xoptarg_c,FEXT_TSV) == 0) {
        format_in = FMT_TSV;
      }
      else {
        fprintf(stderr,
                "Unexpected file format: %s (Use -h for usage)\n",
                xoptarg);
        exit(1);
      }
      break;
    case O_OUTPUT:
      tolower_str(xoptarg_c);
      if (strcmp(xoptarg_c,FEXT_CSV) == 0) {
        format_out = FMT_CSV;
      }
      else if (strcmp(xoptarg_c,FEXT_TSV) == 0) {
        format_out = FMT_TSV;
      }
      else {
        fprintf(stderr,
                "Unexpected file format: %s (Use -h for usage)\n",
                xoptarg);
        exit(1);
      }
      break;
    case O_SUFFIX:
      mode_suffix = 1;
      copy_optarg(fname_suffix,MAX_FILENAME);
      break;
    case O_SAVE_MESSAGE:
      mode_msg = 1;
      break;
    case O_READ_PARAM:
      mode_em = 0;
      break;
    default:
      break;
    }
  }

  build_filename();

  /* after this we can use the die() family and xmalloc() */
  open_log();

  /* Check -A option */
  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_AUTOCLASS:
      smooth_c = parse_smooth_c(SMOOTH_AUTO);
      num_smooth_w = parse_smooth_w(SMOOTH_AUTO);
      tau = parse_smooth_tau(SMOOTH_AUTO);
      ic_type = AUTOCLASS;
      break;
    default:
      break;
    }
  }

  /* Check --n-attr (-x) option, which is referred to by --ignore (-y) option */
  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_N_ATTR:
      tuple_size = parse_int(xoptarg);
      break;
    default:
      break;
    }
  }
  if (tuple_size < 1)
    die("Missing or illegal %s option (Use -h for usage)\n",OPTSTR(O_N_ATTR));

  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_EVAL:
      mode_eval = 1;
      attr_eval = parse_int(xoptarg);
      break;
    case O_ID:
      mode_id = 1;
      attr_id = parse_int(xoptarg);
      break;
    default:
      break;
    }
  }

  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_N_CAND:
      max_candidate = parse_int(xoptarg);
      break;
    case O_DAEM:
      mode_daem = 1;
      break;
    case O_ITEMP:
      parse_itemp(xoptarg);
      break;
    case O_SMOOTH_C:
      smooth_c = parse_smooth_c(xoptarg);
      break;
    case O_MIN_NEW_C_PARAM:
      min_new_param_c = parse_float(xoptarg);
      break;
    case O_DISSIM:
      mode_dissim = 1;
      break;
    case O_MEM_PROB:
      mode_memp = 1;
      break;
    case O_EPSILON:
      epsilon = parse_float(xoptarg);
      break;
    case O_NORM_JMERGE:
      mode_norm = 1;
      break;
    case O_SCORE:
      toupper_str(xoptarg_c); /* working on temporary area */
      if (strcmp(xoptarg_c,"BIC") == 0) {
        ic_type = BIC;
      }
      else if (strcmp(xoptarg_c,"AUTOCLASS") == 0 ||
               strcmp(xoptarg_c,"CS") == 0) {
        ic_type = AUTOCLASS;
      }
      else {
        fprintf(stderr,
                "Unexpected information criterion: %s (Use -h for usage)\n",
                xoptarg);
        exit(1);
      }
      break;
    case O_N_JSPLIT:
      max_jsplit = parse_int(xoptarg);
      break;
    case O_READ_HPARAM:
      mode_read_hparam = 1;
      break;
    case O_INIT_C:
      tolower_str(xoptarg_c);
      if (strcmp(xoptarg_c,INIT_STR_RANDOM) == 0)
        init_type_class = RANDOM_C;
      else if (strcmp(xoptarg_c,INIT_STR_NOISY_U) == 0)
        init_type_class = NOISY_UNIFORM_C;
      else
        die("Unknown initialization method -- %s (Use -h for usage)\n",xoptarg);
      break;
    case O_INIT_A:
      tolower_str(xoptarg_c);
      if (strcmp(xoptarg_c,INIT_STR_RANDOM) == 0)
        init_type_word = RANDOM_W;
      else if (strcmp(xoptarg_c,INIT_STR_NOISY_U) == 0)
        init_type_word = NOISY_UNIFORM_W;
      else if (strcmp(xoptarg_c,INIT_STR_VFREQ) == 0)
        init_type_word = WORD_FREQ;
      else
        die("Unknown initialization method -- %s (Use -h for usage)\n",xoptarg);
      break;
    case O_N_CLASS:
      parse_num_class(xoptarg);  /* we need a special care */
      break;
    case O_SMOOTH_MEAN:
      smooth_u = parse_float(xoptarg);
      mode_smooth_u = 1;
      break;
    case O_MAX_ITER_RESTART:
      max_init_iterate = parse_int(xoptarg);
      break;
    case O_SMOOTH_VAR:
      parse_smooth_s(xoptarg);
      break;
    case O_MAX_ITER:
      max_iterate = parse_int(xoptarg);
      break;
    case O_RESTART:
      num_init = parse_int(xoptarg);
      break;
    case O_MIN_SM_IMP:
      min_smem_improve = parse_float(xoptarg);
      break;
    case O_N_SM:
      num_sm = parse_int(xoptarg);
      break;
    case O_ALPHA:
      alpha_w = parse_float(xoptarg);
      break;
#ifdef USE_ICONV
    case O_CODE_IN:
      copy_optarg(charcode_in,CHARCODE_LEN);
      break;
#endif
#ifdef USE_ICONV
    case O_CODE_OUT:
      copy_optarg(charcode_out,CHARCODE_LEN);
      break;
#endif
    case O_SEED:
      rand_seed = parse_uint(xoptarg);
      mode_set_seed = 1;
      break;
    case O_TAU:
      tau = parse_smooth_tau(xoptarg);
      break;
    case O_GAUSS:
      mode_gauss = 1;
      add_gauss(xoptarg);
      break;
    case O_MIN_C_PARAM:
      min_param_c = parse_float(xoptarg);
      break;
    case O_VERBOSE:
      mode_verbose = 1;
      break;
    case O_SMOOTH_A:
      num_smooth_w = parse_smooth_w(xoptarg);
      break;
    case O_OCC_ORDER:
      mode_lex_order = 0;
      break;
    case O_RANK:
      mode_rank = 1;
      parse_rank(xoptarg);
      break;
    case O_RANK_HIDE:
      mode_rank_hide = 1;
      break;
    case O_LOG_SCALE:
      mode_log_scale = 1;
      break;
    case O_IGNORE:
      mode_ignore = 1;
      add_attr_ignore(xoptarg);
      break;
    case O_HIDE_ATTR:
      mode_hide_attr = 1;
      break;
    case O_PREC_OUT:
      prec_out = parse_int(xoptarg);
      break;
    case O_PREC_MSG:
      prec_msg = parse_int(xoptarg);
      break;
    case O_UNCOMPRESS:
      mode_uncompress = 1;
      break;
    case O_W_PURITY:
      weight_purity = parse_float(xoptarg);
      break;
    case O_W_WALLANCE:
      weight_wallance = parse_float(xoptarg);
      break;
    case O_SHOW_ATTR_MEMP:
      mode_show_attr_memp = 1;
      break;
    case O_SHOW_ATTR_CLUSTER:
      mode_show_attr_cluster = 1;
      break;
    case O_NOISE_RATIO:
      noise_ratio = parse_float(xoptarg);
      break;
    case O_NOISE_RATIO_GAUSS:
      noise_ratio_gauss = parse_float(xoptarg);
      break;
    case O_EXTREME:
      parse_extreme(xoptarg);
      break;
    case O_REGULAR_STAT:
      mode_regular_stat = 1;
      break;
    case O_NOOP:
      /* do nothing */
      break;
    default:
      break;
    }
  }
}

static void postproc_args(void)
{
  int v,u;

  check_tuple_size();

  /* tuple_size_d and mode_nominal are computed here */
  build_attr_info();

  if (mode_nominal) build_smooth_nominal(num_smooth_w);

  /* default settings for hyperparameters of Gaussians */
  if (mode_gauss) {
    if (tau > 0.0 || mode_smooth_auto_tau == MODE_SMOOTH_AUTO) {
      mode_smooth_gauss = 1;
      for (u=0; u<gauss_size; u++) {
        if (alpha_w < 0 || (alpha_w > 0 && alpha_w <= gauss_attr_size[u]))
          die("Missing or illegal value for %s option\n",OPTSTR(O_ALPHA));
      }
    }
    else if (tau < 0.0)
      die("Illegal value for %s option\n",OPTSTR(O_TAU));
  }

  if (mode_id && (attr_id < 0 || attr_id >= tuple_size))
    die("Illegal value for %s option\n",OPTSTR(O_ID));

  if (mode_eval && (attr_eval < 0 || attr_eval >= tuple_size))
    die("Illegal value for %s option\n",OPTSTR(O_EVAL));

  if (mode_id && mode_eval && attr_id == attr_eval)
    die("Confliction between %s and %s options\n",OPTSTR(O_ID),OPTSTR(O_EVAL));

  if (mode_ignore) {
    num_attr_ignore = 0;
    for (v=0; v<tuple_size; v++) {
      if (attr_ignore[v]) {
        if (v == attr_id)
          die("Confliction between %s and %s options\n",
              OPTSTR(O_ID),OPTSTR(O_IGNORE));
        else if (v == attr_eval)
          die("Confliction between %s and %s options\n",
              OPTSTR(O_IGNORE),OPTSTR(O_EVAL));
        else
          num_attr_ignore++;
      }
    }
  }

  if (mode_em && mode_search_num_class) {
    if (kmin <= 1 || kmax <= 0 || kstep <= 0 || kmin > kmax)
      die("Illegal value for %s option (Use -h for usage)\n",
          OPTSTR(O_N_CLASS));
  }
  else if (mode_em && !mode_read_hparam && num_class <= 1) {
    die("Missing or illegal %s option (Use -h for usage)\n",
        OPTSTR(O_N_CLASS));
  }

  /* In the later versions, this will be more complicated */
  mode_as_is = (mode_em) ? 1 : 0;

  if (max_iterate < 0 || max_init_iterate < 0)
    die("Illegal value for the number of iterations\n");

  if (max_iterate > MAX_MAX_ITERATE)
    die("Maximum number of iterations cannot exceed %d\n",MAX_MAX_ITERATE);

  if (num_init <= 0)
    die("Illegal value for the number of initial parameter sets\n");

  if (max_jsplit < 0)
    die("Illegal value for the number of split candidate to be examined\n");

  if (min_param_c < 0)
    die("Illegal value for the lower limit of prob. of the class"
        " to be split/merge\n");

  if (min_new_param_c <= 0.0)
    die("Illegal value for the lower limit of updated prob. of"
        " the splited/merged class\n");

  if (prec_out <= 0 || prec_out > MAX_PREC_OUT)
    die("Illegal value for %s option\n",OPTSTR(O_PREC_OUT));

  if (prec_msg <= 0 || prec_msg > MAX_PREC_MSG)
    die("Illegal value for %s option\n",OPTSTR(O_PREC_MSG));

#ifdef USE_ICONV
  /* set the mode for character code conversion */
  mode_conv_in  = (strcmp(charcode_in, CHARCODE_INTERNAL) == 0) ? 0 : 1;
  mode_conv_out = (strcmp(charcode_out,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#else
  mode_conv_in  = 0;
  mode_conv_out = 0;
#endif

  /* get a string representation of the model score */
  switch (ic_type) {
  case BIC:
    strncpy(ic_type_str,IC_STR_BIC,ENUM_TYPE_LEN);
    break;
  case AUTOCLASS:
    strncpy(ic_type_str,IC_STR_AUTOCLASS,ENUM_TYPE_LEN);
    break;
  default:
    die("Internal error: Illegal Info. Criterion\n");
  }

  /* get a string representation of the clustering score */
  if (mode_cluster) {
    switch (cl_type) {
    case CL_MP:
      strncpy(cl_type_str,"MEMBERSHIP_DIST",ENUM_TYPE_LEN);
      break;
    default:
      die("Internal error: Illegal Info. Criterion\n");
    }
  }

  /* get a string representation of initialization methods for class params */
  switch (init_type_class) {
  case NOISY_UNIFORM_C:
    strncpy(init_type_str_class,"NOISY_UNIFORM",ENUM_TYPE_LEN);
    break;
  case RANDOM_C:
    strncpy(init_type_str_class,"RANDOM",ENUM_TYPE_LEN);
    break;
  default:
    die("Internal error: Illegal Info. Criterion\n");
  }

  /* get a string representation of initialization methods for attr. params */
  switch (init_type_word) {
  case NOISY_UNIFORM_W:
    strncpy(init_type_str_word,"NOISY_UNIFORM",ENUM_TYPE_LEN);
    break;
  case RANDOM_W:
    strncpy(init_type_str_word,"RANDOM",ENUM_TYPE_LEN);
    break;
  case WORD_FREQ:
    strncpy(init_type_str_word,"WORD_FREQ",ENUM_TYPE_LEN);
    break;
  default:
    die("Internal error: Illegal initialization type for class\n");
  }

  /* get a string representation of the relevance scores */
  switch (rel_m) {
  case REL_MP:
    strncpy(rel_m_str,"MEMBERSHIP_DIST",ENUM_TYPE_LEN);
    break;
  default:
    die("Internal error: Illegal initialization type for attributes\n");
  }

  /* Set the lower limit of the improvement by a split-merge operation
   * (Note that the default value depends on the '-e' flag). */
  if (min_smem_improve < 0) {
    min_smem_improve = epsilon * MIN_SMEM_IMPROVE_RATIO;
  }

  /* Get the number of inverse temperature to be examined */
  if (mode_daem)
    num_itemp = (int)(-LOG(itemp_init)/LOG(itemp_rate)) + 1;

  /* Check the weights for F-measures */
  if (mode_eval) {
    if (weight_purity < 0)
      die("Illegal value for %s option\n",OPTSTR(O_W_PURITY));
    if (weight_wallance < 0)
      die("Illegal value for %s option\n",OPTSTR(O_W_WALLANCE));
  }
}

int main(int argc, char *argv[])
{
  int v;

  initialize_mode();
  initialize_buffer();
  initialize_printer();
#ifdef USE_ICONV
  initialize_iconv_sys();
#endif
  initialize_matrix_pool();
  initialize_vector_pool();

  parse_args(argc,argv);
  postproc_args();

#ifdef USE_ICONV
  initialize_iconv();
#endif
  initialize_rand();

  print_version(fp_msg);
  print_settings(fp_log);

  if (mode_em) em_main();
  else no_em_main();

  fclose(fp_log);
#ifdef USE_ICONV
  iconv_close(cd_in);
  iconv_close(cd_out);
#endif

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      release_hashtable(&ht[v]);
  }

  release_matrix_pool();
  release_vector_pool();

  release_buffer();
  release_printer();

  if (smooth_w != NULL) xfree(smooth_w);

  return 0;
}
