/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#ifdef VNBC
#  include "vnbc_def.h"
#  include "vnbc_ext.h"
#endif

static FLOAT_T *tmp_param_c;
static FLOAT_T *tmp_rel_m;

static int compare_prob_mem(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  FLOAT_T px = tmp_rel_m[ix];
  FLOAT_T py = tmp_rel_m[iy];

  if (px < py)  return 1;
  if (px > py)  return -1;

  return 0;
}

static int compare_param_c(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  FLOAT_T px = tmp_param_c[ix];
  FLOAT_T py = tmp_param_c[iy];

  if (px < py) return 1;
  if (px > py) return -1;

  return 0;
}

static void compute_rank_case(void)
{
  int k,e;

  for (k=0; k<num_class; k++)
    rank_class[k] = k;

  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      rank_mem_case[k][e] = e;

  tmp_param_c = param_c;
  qsort(rank_class,num_class,sizeof(int),
        (int(*)(const void *,const void *))compare_param_c);

  switch (rel_m) {
  case REL_MP:
    for (k=0; k<num_class; k++) {
      tmp_rel_m = prob_mem_case[k];
      qsort(rank_mem_case[k],num_entry,sizeof(int),
            (int(*)(const void *, const void *))compare_prob_mem);
    }
    break;
  default:
    die("Internal error:"
        " Unexpected relevance measure between classes and attribute values\n");
  }
}

static void compute_rank_w(void)
{
  int k,v,i;

  for (k=0; k<num_class; k++)
    rank_class[k] = k;

  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size_d; v++)
      for (i=0; i<num_word[v]; i++)
        rank_mem_w[k][v][i] = i;

  tmp_param_c = param_c;
  qsort(rank_class,num_class,sizeof(int),
        (int(*)(const void *,const void *))compare_param_c);

  switch (rel_m) {
  case REL_MP:
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        tmp_rel_m = prob_mem_w[k][v];
        qsort(rank_mem_w[k][v],num_word[v],sizeof(int),
              (int(*)(const void *, const void *))compare_prob_mem);
      }
    }
    break;
  default:
    die("Internal error:"
        " Unexpected relevance measure between classes and attribute values\n");
  }
}

static void compute_rank_g(void)
{
  int k,u,e;

  for (k=0; k<num_class; k++)
    rank_class[k] = k;

  for (k=0; k<num_class; k++)
    for (u=0; u<gauss_size; u++)
      for (e=0; e<num_entry; e++)
        rank_mem_g[k][u][e] = e;

  tmp_param_c = param_c;
  qsort(rank_class,num_class,sizeof(int),
        (int(*)(const void *,const void *))compare_param_c);

  switch (rel_m) {
  case REL_MP:
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        tmp_rel_m = prob_mem_g[k][u];
        qsort(rank_mem_g[k][u],num_entry,sizeof(int),
              (int(*)(const void *, const void *))compare_prob_mem);
      }
    }
    break;
  default:
    die("Internal error:"
        " Unexpected relevance measure between classes and attribute values\n");
  }
}

void compute_rank(void)
{
  compute_rank_case();
  if (mode_nominal) compute_rank_w();
  if (mode_gauss) compute_rank_g();
}


void rank_main(void)
{
  FILE *fp_rank;

  print_msg("Ranking relevant cases/attvals...");
  fflush(fp_msg);

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_rank = current_sec();
  stime_lap_rank = elapsed_sec();
#endif

  alloc_rank();
  compute_rank();

  fp_rank = fopen(fname_rank,"wb");
  if (fp_rank == NULL) die("Couldn't open %s\n",fname_rank);

  print_rank(fp_rank);
  fclose(fp_rank);

  release_rank();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_rank = current_sec();
  etime_lap_rank = elapsed_sec();
#endif

  print_msg("done\n");
}
